# -*- coding: utf-8 -*-
# Copyright (C) 2010-2018 Linaro Limited
#
# Author: Zygmunt Krynicki <zygmunt.krynicki@linaro.org>
#         Michael Hudson-Doyle <michael.hudson@linaro.org>
#
# This file is part of LAVA.
#
# LAVA is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License version 3
# as published by the Free Software Foundation
#
# LAVA is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with LAVA.  If not, see <http://www.gnu.org/licenses/>.

import os
import subprocess  # nosec internal
import xmlrpc.client
import yaml

from django.http import Http404
from django.conf import settings
from django.core.exceptions import PermissionDenied

from lava_common.utils import debian_package_version
from lava_scheduler_app.views import get_restricted_job
from lava_scheduler_app.models import Device, DeviceType
from linaro_django_xmlrpc.models import errors, Mapper, SystemAPI


class LavaSystemAPI(SystemAPI):
    """
    Extend the default SystemAPI with a 'whoami' method.
    """

    def whoami(self):
        """
        Name
        ----
        `whoami` ()

        Description
        -----------
        Find the authenticated user, if any, or None.

        Arguments
        ---------
        None

        Return value
        ------------
        Name of the authenticated user, if any, or None.
        """
        if self.user:
            return self.user.username
        else:
            return None

    def version(self):
        """
        Name
        ----
        `system.version` ()

        Description
        -----------
        Return the lava-server version string

        Arguments
        ---------
        None

        Return value
        ------------
        lava-server version string
        """
        return debian_package_version(pkg="lava-server", split=False)

    # Update the integer return value when adding arguments to
    # existing functions anywhere in the XML-RPC API or
    # deleting functions.
    def api_version(self):
        """
        Name
        ----
        `system.api_version` ()

        Description
        -----------
        Return the lava-server XML-RPC API version string.
        Clients can check this string to know whether to
        use particular arguments to available functions.

        Note: For older instances which do not support this
        call in the first place, check for this call in the
        output of system.listMethods()

        if 'system.api_version' in connection.system.listMethods():
            api_version = int(connection.system.api_version())
                if api_version >= 2:
                    # safe to run the new API here.
        else:
            # use the old API

        Arguments
        ---------
        None

        Return value
        ------------
        lava-server XML-RPC API version integer
        """
        return 2

    def master_config(self):
        """
        Name
        ----
        `master_config` ()

        Description
        -----------
        Return a dictionary containing the master and logger ZMQ
        socket addresses for this instance.

        Arguments
        ---------
        None

        Return value
        ------------
        Returns a dictionary containing the following keys:
        {
          "MASTER_URL": "tcp://<lava-master-dns>:5556",
          "LOGGING_URL": "tcp://<lava-master-dns>:5555",
          "ENCRYPT": False,
          "IPv6": False,
          "EVENT_SOCKET": "tcp://*:5500",
          "EVENT_TOPIC": "org.linaro.validation",
          "EVENT_NOTIFICATION": True,
          "LOG_SIZE_LIMIT": 10,
        }

        If ENCRYPT is True, clients MUST already have a usable
        client certificate installed on the master AND the current
        master certificate installed on the client, before a
        connection can be made.
        """
        data = {
            "master_socket": "tcp://<lava-master-dns>:5556",
            "socket": "tcp://<lava-master-dns>:5555",
            "encrypt": False,
            "ipv6": False,
        }

        master = {"ERROR": "invalid master config"}
        filename = os.path.join(settings.MEDIA_ROOT, "lava-master-config.yaml")
        if os.path.exists(filename):
            try:
                with open(filename, "r") as output:
                    master = yaml.safe_load(output)
            except yaml.YAMLError as exc:
                return master
        if master:
            data.update(master)

        log_config = {"ERROR": "invalid logging config"}
        filename = os.path.join(settings.MEDIA_ROOT, "lava-logs-config.yaml")
        if os.path.exists(filename):
            try:
                with open(filename, "r") as output:
                    log_config = yaml.safe_load(output)
            except yaml.YAMLError as exc:
                return log_config
        if log_config:
            data.update(log_config)
        return {
            "MASTER_URL": data["master_socket"],
            "LOGGING_URL": data["socket"],
            "IPv6": data["ipv6"],
            "ENCRYPT": data.get("encrypt", False),
            "EVENT_TOPIC": settings.EVENT_TOPIC,
            "EVENT_SOCKET": settings.EVENT_SOCKET,
            "EVENT_NOTIFICATION": settings.EVENT_NOTIFICATION,
            "LOG_SIZE_LIMIT": settings.LOG_SIZE_LIMIT,
        }

    def user_can_view_jobs(self, job_list, username=None):
        """
        Name
        ----
        user_can_view_jobs (`job_list`)

        Administrators only:
        user_can_view_jobs (`job_list`, `username`)

        Description
        -----------
        Check the access permissions on a list of jobs.
        Admins can specify a username as the second argument to run
        the query on behalf of that user.

        Arguments
        ---------
        job_list: list
            list of job ids to query
        username: string
            username of the user to query (admins only)

        Return value
        ------------
        Returns a dictionary where the key is a string of the job_id from
        the job_list, if it exists in the queried instance. The value is a boolean
        for whether the user can access that job.
        {
          '1234': True,
          '1543': False
        }
        If the job number does not exist, that job will be omitted from the
        returned dictionary.
        This function requires authentication with a username and token.

        Example
        -------
        server.system.user_can_view_jobs([1, 2, 3, 4, 99999, 4000])
        {'1': True, '4000': True, '3': True, '2': True, '4': True}

        # if using the username and token of an admin, a different user can be queried:
        server.system.user_can_view_jobs([1, 2, 3, 4, 99999, 4000], 'firstname.lastname')
        {'1': True, '4000': False, '3': True, '2': True, '4': False}

        """
        self._authenticate()
        if not isinstance(job_list, list):
            raise xmlrpc.client.Fault(
                errors.BAD_REQUEST, "job list argument must be a list"
            )
        username = self._switch_user(username)
        retval = {}
        for job_id in job_list:
            try:
                get_restricted_job(username, job_id)
            except Http404:
                continue
            except PermissionDenied:
                retval[str(job_id)] = False
                continue
            retval[str(job_id)] = True
        return retval

    def user_can_view_bundles(self, bundle_list, username=None):
        """
        Name
        ----
        user_can_view_bundles (`bundle_list`)

        Removal of V1 support
        --------------------
        This function has been disabled. It is retained as a stub for older
        versions of clients. Please update your tool to use LAVA V2.

        """
        return False

    def user_can_view_devices(self, device_list, username=None):
        """
        Name
        ----
        user_can_view_devices (`device_list`)

        Administrators only:
        user_can_view_devices (`device_list`, `username`)

        Description
        -----------
        Check the access permissions on a list of devices.
        Admins can specify a username as the second argument to run
        the query on behalf of that user.

        Arguments
        ---------
        device_list: list
            list of device hostnames to query
        username: string
            username of the user to query (admins only)

        Return value
        ------------
        Returns a nested dictionary where the top level key is the device type, the value is
        a list of dictionaries. The key of the second dictionary is a hostname from the device_list
        which exists in the queried instance. The value is a boolean
        for whether the user can access that device.
        If the device is visible, also includes a boolean denoting whether the specified
        device is a pipeline device which can run jobs designed in the dispatcher refactoring.
        If a pipeline device, also includes a boolean denoting whether the specified pipeline
        device exclusively accepts YAML submissions - JSON submissions are rejected if this
        device is marked as exclusive.
        {
            'mustang': [
                {
                    'mustang01': {
                        'visible': True
                        }
                }
            ],
            'panda': [
                {
                    'panda05': {
                        'visible': True
                        }
                    }
            ]
        }

        If the device type is hidden, that type and the nested dictionary for that type
        will be omitted.
        Retired devices will be omitted.
        If the device does not exist, that device will be omitted from the
        returned dictionary.
        This function requires authentication with a username and token.

        Example
        -------

        server.system.user_can_view_devices(['kvm014', 'kvm02'])
        {'qemu':
            [
                {'kvm014': {
                    'exclusive': True,
                    'visible': True,
                    'is_pipeline': True
                    }
                },
                {'kvm02': {
                    'exclusive': True,
                    'visible': True,
                    'is_pipeline': True}
                }
            ]
        }

        # if using the username and token of an admin, a different user can be queried:
        server.system.user_can_view_devices(
            ['mustang01', 'wandboard07', 'imx53-01', 'cubietruck02', 'black01'],
            'firstname.lastname'
        )

        """
        self._authenticate()
        if not isinstance(device_list, list):
            raise xmlrpc.client.Fault(
                errors.BAD_REQUEST, "device list argument must be a list"
            )
        username = self._switch_user(username)
        retval = {}
        for hostname in device_list:
            try:
                device = Device.objects.get(hostname=hostname)
            except Device.DoesNotExist:
                continue
            device_type = device.device_type
            if device_type.owners_only and not device.is_owned_by(username):
                continue
            retval.setdefault(device_type.name, [])
            visible = device.is_visible_to(username)
            if visible:
                retval[device_type.name].append(
                    {
                        hostname: {
                            "is_pipeline": True,
                            "visible": visible,
                            "exclusive": True,
                        }
                    }
                )
            else:
                retval[device_type.name].append({hostname: {"visible": visible}})
        return retval

    def user_can_submit_to_types(self, type_list, username=None):
        """
        Name
        ----
        user_can_submit_to_types (`type_list`)

        Administrators only:
        user_can_submit_to_types (`type_list`, `username`)

        Description
        -----------
        Check the access permissions on a list of device types.
        Admins can specify a username as the second argument to run
        the query on behalf of that user.

        Arguments
        ---------
        type_list: list
            list of device types to query
        username: string
            username of the user to query (admins only)

        Return value
        ------------
        Returns a dictionary where the key is a string of the device_type from
        the type_list, if it exists in the queried instance. The value is a boolean
        for whether the user can submit to any devices of the specified type.
        {
          'panda': True,
          'mustang': False
        }
        If the device type number does not exist, that type will be omitted from the
        returned dictionary.
        This function requires authentication with a username and token.

        Example
        -------
        server.system.user_can_submit_to_types(
            ['mustang', 'panda', 'qemu', 'kvm', 'cubie2']
        )
        {'cubie2': True, 'mustang': False, 'kvm': True, 'qemu': True, 'panda': True}

        # if using the username and token of an admin, a different user can be queried:
        server.system.user_can_submit_to_types(
            ['mustang', 'panda', 'qemu', 'kvm', 'cubie2'],
            'firstname.lastname'
        )

        """
        self._authenticate()
        if not isinstance(type_list, list):
            raise xmlrpc.client.Fault(
                errors.BAD_REQUEST, "type list argument must be a list"
            )
        username = self._switch_user(username)
        if not username.has_perm("lava_scheduler_app.add_testjob"):
            raise xmlrpc.client.Fault(
                errors.FORBIDDEN,
                "User '%s' does not have permissiont to submit jobs." % username,
            )
        retval = {}
        for type_name in type_list:
            try:
                device_type = DeviceType.objects.get(name=type_name)
            except DeviceType.DoesNotExist:
                continue
            devices = Device.objects.filter(device_type=device_type)
            access = []
            for device in devices:
                access.append(device.can_submit(username))
            retval[device_type.name] = any(access)
        return retval

    def pipeline_network_map(self, switch=None):
        """
        Name
        ----
        pipeline_network_map(switch=None):

        Description
        -----------

        Collate all the vland information from pipeline devices to create a complete map,
        then return YAML data for all switches or a specified switch.

        This function requires authentication with a username and token.

        Arguments
        ---------
        switch: string
            hostname or IP of the switch to map. If None, data is returned for all switches.

        Return value
        ------------
        Returns a YAML file of the form:

        switches:
          '192.168.0.2':
          - port: 5
            device:
              interface: eth0
              sysfs: "/sys/devices/pci0000:00/0000:00:19.0/net/eth0"
              mac: "f0:de:f1:46:8c:21"
              hostname: bbb1

        """
        self._authenticate()
        # get all device dictionaries, build the entire map.
        dictionaries = [
            (device.hostname, device.load_configuration())
            for device in Device.objects.all()
        ]
        network_map = {"switches": {}}
        for (hostname, params) in dictionaries:
            if "interfaces" not in params:
                continue
            for interface in params["interfaces"]:
                for map_switch, port in params["map"][interface].items():
                    port_list = []
                    device = {
                        "interface": interface,
                        "mac": params["mac_addr"][interface],
                        "sysfs": params["sysfs"][interface],
                        "hostname": hostname,
                    }
                    port_list.append({"port": port, "device": device})
                    switch_port = network_map["switches"].setdefault(map_switch, [])
                    # Any switch can only have one entry for one port
                    if port not in switch_port:
                        switch_port.extend(port_list)

        if switch:
            if switch in network_map["switches"]:
                return yaml.dump(network_map["switches"][switch])
            else:
                return xmlrpc.client.Fault(
                    404,
                    "No switch '%s' was found in the network map of supported devices."
                    % switch,
                )
        return yaml.dump(network_map)


class LavaMapper(Mapper):
    def register_introspection_methods(self):
        """
        Register LavaSystemAPI as 'system' object.

        LavaSystemAPI adds a 'whoami' system method above what the default
        has.
        """
        self.register(LavaSystemAPI, "system")



# Copyright (C) 2015 Linaro Limited
#
# Author: Neil Williams <neil.williams@linaro.org>
#
# This file is part of LAVA Dispatcher.
#
# LAVA Dispatcher is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# LAVA Dispatcher is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along
# with this program; if not, see <http://www.gnu.org/licenses>.


import os
import yaml
import socket
from lava_dispatcher.device import NewDevice
from lava_dispatcher.parser import JobParser
from lava_common.exceptions import JobError
from lava_dispatcher.connection import Protocol
from lava_dispatcher.protocols.vland import VlandProtocol
from lava_dispatcher.protocols.multinode import MultinodeProtocol
from lava_dispatcher.tests.test_basic import StdoutTestCase, Factory
from lava_dispatcher.tests.utils import DummyLogger

# pylint: disable=superfluous-parens


class TestVland(StdoutTestCase):  # pylint: disable=too-many-public-methods
    def setUp(self):
        super().setUp()
        self.filename = os.path.join(
            os.path.dirname(__file__), "sample_jobs/bbb-group-vland-alpha.yaml"
        )
        self.beta_filename = os.path.join(
            os.path.dirname(__file__), "sample_jobs/bbb-group-vland-beta.yaml"
        )
        self.factory = Factory()
        (rendered, _) = self.factory.create_device("bbb-01.jinja2")
        self.device = NewDevice(yaml.safe_load(rendered))
        self.job_id = "100"

    def test_file_structure(self):
        with open(self.filename) as yaml_data:
            alpha_data = yaml.safe_load(yaml_data)
        self.assertIn("protocols", alpha_data)
        self.assertTrue(VlandProtocol.accepts(alpha_data))
        level_tuple = Protocol.select_all(alpha_data)
        self.assertEqual(len(level_tuple), 2)
        self.assertEqual(
            VlandProtocol,
            [item[0] for item in sorted(level_tuple, key=lambda data: data[1])][1],
        )
        vprotocol = VlandProtocol(alpha_data, self.job_id)
        self.assertIn("arbit", vprotocol.base_group)
        self.assertNotIn("group", vprotocol.base_group)
        vprotocol.set_up()
        self.assertIn("port", vprotocol.settings)
        self.assertIn("poll_delay", vprotocol.settings)
        self.assertIn("vland_hostname", vprotocol.settings)
        self.assertEqual(
            vprotocol.base_message,
            {
                "port": vprotocol.settings["port"],
                "poll_delay": vprotocol.settings["poll_delay"],
                "host": vprotocol.settings["vland_hostname"],
                "client_name": socket.gethostname(),
            },
        )
        for name in vprotocol.names:
            vlan = vprotocol.params[name]
            self.assertIn("tags", vlan)

    def test_device(self):
        self.assertIsNotNone(self.device)
        self.assertIn("eth0", self.device["parameters"]["interfaces"])
        self.assertIn("eth1", self.device["parameters"]["interfaces"])
        self.assertIn("sysfs", self.device["parameters"]["interfaces"]["eth0"])
        self.assertIn("mac", self.device["parameters"]["interfaces"]["eth0"])
        self.assertIn("switch", self.device["parameters"]["interfaces"]["eth0"])
        self.assertIn("port", self.device["parameters"]["interfaces"]["eth0"])
        self.assertIn("tags", self.device["parameters"]["interfaces"]["eth0"])
        self.assertIn("sysfs", self.device["parameters"]["interfaces"]["eth1"])
        self.assertIn("mac", self.device["parameters"]["interfaces"]["eth1"])
        self.assertIn("switch", self.device["parameters"]["interfaces"]["eth1"])
        self.assertIn("port", self.device["parameters"]["interfaces"]["eth1"])
        self.assertIn("tags", self.device["parameters"]["interfaces"]["eth1"])
        self.assertIsInstance(
            self.device["parameters"]["interfaces"]["eth1"]["tags"], list
        )
        self.assertIsNone(self.device["parameters"]["interfaces"]["eth0"]["tags"])
        csv_list = []
        for interface in self.device["parameters"]["interfaces"]:
            csv_list.extend(
                [
                    self.device["parameters"]["interfaces"][interface]["sysfs"],
                    self.device["parameters"]["interfaces"][interface]["mac"],
                    interface,
                ]
            )
        self.assertEqual(
            set(csv_list),
            {
                "/sys/devices/pci0000:00/0000:00:1c.1/0000:03:00.0/net/eth1",
                "00:24:d7:9b:c0:8c",
                "eth1",
                "/sys/devices/pci0000:00/0000:00:19.0/net/eth0",
                "f0:de:f1:46:8c:21",
                "eth0",
            },
        )
        tag_list = []
        for interface in self.device["parameters"]["interfaces"]:
            if interface == "eth0":
                continue
            for tag in self.device["parameters"]["interfaces"][interface]["tags"]:
                tag_list.extend([interface, tag])
        self.assertEqual(set(tag_list), {"RJ45", "100M", "eth1", "10M"})

    def test_configure(self):
        with open(self.filename) as yaml_data:
            alpha_data = yaml.safe_load(yaml_data)
        self.assertIn("protocols", alpha_data)
        self.assertTrue(VlandProtocol.accepts(alpha_data))
        vprotocol = VlandProtocol(alpha_data, self.job_id)
        vprotocol.set_up()
        with open(self.filename) as sample_job_data:
            parser = JobParser()
            job = parser.parse(sample_job_data, self.device, 4212, None, "")
        ret = vprotocol.configure(self.device, job)
        if not ret:
            print(vprotocol.errors)
        self.assertTrue(ret)
        nodes = {}
        for name in vprotocol.names:
            vlan = vprotocol.params[name]
            # self.assertNotIn('tags', vlan)
            uid = " ".join([vlan["switch"], str(vlan["port"])])
            nodes[uid] = name
        self.assertEqual(len(nodes.keys()), len(vprotocol.names))
        self.assertIn("vlan_one", vprotocol.names)
        self.assertNotIn("vlan_two", vprotocol.names)
        self.assertIn("switch", vprotocol.params["vlan_one"])
        self.assertIn("port", vprotocol.params["vlan_one"])
        self.assertIsNotNone(vprotocol.multinode_protocol)

        (rendered, _) = self.factory.create_device("bbb-01.jinja2")
        bbb2 = NewDevice(yaml.safe_load(rendered))
        bbb2["parameters"]["interfaces"]["eth0"]["switch"] = "192.168.0.2"
        bbb2["parameters"]["interfaces"]["eth0"]["port"] = "6"
        bbb2["parameters"]["interfaces"]["eth1"]["switch"] = "192.168.0.2"
        bbb2["parameters"]["interfaces"]["eth1"]["port"] = "4"
        self.assertEqual(
            vprotocol.params,
            {
                "vlan_one": {
                    "switch": "192.168.0.2",
                    "iface": "eth1",
                    "port": 7,
                    "tags": ["100M", "RJ45", "10M"],
                }
            },
        )
        # already configured the vland protocol in the same job
        self.assertTrue(vprotocol.configure(bbb2, job))
        self.assertEqual(
            vprotocol.params,
            {
                "vlan_one": {
                    "switch": "192.168.0.2",
                    "iface": "eth1",
                    "port": 7,
                    "tags": ["100M", "RJ45", "10M"],
                }
            },
        )
        self.assertTrue(vprotocol.valid)
        self.assertEqual(vprotocol.names, {"vlan_one": "4212vlanone"})

    def test_job(self):
        with open(self.filename) as yaml_data:
            alpha_data = yaml.safe_load(yaml_data)
        self.assertIn("protocols", alpha_data)
        self.assertIn(VlandProtocol.name, alpha_data["protocols"])
        with open(self.filename) as sample_job_data:
            parser = JobParser()
            job = parser.parse(sample_job_data, self.device, 4212, None, "")
        job.logger = DummyLogger()
        description_ref = self.pipeline_reference("bbb-group-vland-alpha.yaml", job=job)
        self.assertEqual(description_ref, job.pipeline.describe(False))
        job.validate()
        self.assertNotEqual(
            [],
            [
                protocol.name
                for protocol in job.protocols
                if protocol.name == MultinodeProtocol.name
            ],
        )
        ret = {
            "message": {"kvm01": {"vlan_name": "name", "vlan_tag": 6}},
            "response": "ack",
        }
        self.assertEqual(
            ("name", 6),
            (ret["message"]["kvm01"]["vlan_name"], ret["message"]["kvm01"]["vlan_tag"]),
        )
        self.assertIn("protocols", job.parameters)
        self.assertIn(VlandProtocol.name, job.parameters["protocols"])
        self.assertIn(MultinodeProtocol.name, job.parameters["protocols"])
        vprotocol = [
            vprotocol
            for vprotocol in job.protocols
            if vprotocol.name == VlandProtocol.name
        ][0]
        self.assertTrue(vprotocol.valid)
        self.assertEqual(vprotocol.names, {"vlan_one": "4212vlanone"})
        self.assertFalse(vprotocol.check_timeout(120, {"request": "no call"}))
        self.assertRaises(JobError, vprotocol.check_timeout, 60, "deploy_vlans")
        self.assertRaises(
            JobError, vprotocol.check_timeout, 60, {"request": "deploy_vlans"}
        )
        self.assertTrue(vprotocol.check_timeout(120, {"request": "deploy_vlans"}))
        for vlan_name in job.parameters["protocols"][VlandProtocol.name]:
            self.assertIn(vlan_name, vprotocol.params)
            self.assertIn("switch", vprotocol.params[vlan_name])
            self.assertIn("port", vprotocol.params[vlan_name])
            self.assertIn("iface", vprotocol.params[vlan_name])
        params = job.parameters["protocols"][vprotocol.name]
        names = []
        for key, _ in params.items():
            names.append(",".join([key, vprotocol.params[key]["iface"]]))
        # this device only has one interface with interface tags
        self.assertEqual(names, ["vlan_one,eth1"])

    def test_vland_overlay(self):
        with open(self.filename) as yaml_data:
            alpha_data = yaml.safe_load(yaml_data)
        for vlan_key, _ in alpha_data["protocols"][VlandProtocol.name].items():
            alpha_data["protocols"][VlandProtocol.name][vlan_key] = {"tags": []}
        # removed tags from original job to simulate job where any interface tags will be acceptable
        self.assertEqual(
            alpha_data["protocols"][VlandProtocol.name], {"vlan_one": {"tags": []}}
        )
        parser = JobParser()
        job = parser.parse(yaml.dump(alpha_data), self.device, 4212, None, "")
        job.logger = DummyLogger()
        job.validate()
        tftp_deploy = [
            action for action in job.pipeline.actions if action.name == "tftp-deploy"
        ][0]
        prepare = [
            action
            for action in tftp_deploy.internal_pipeline.actions
            if action.name == "prepare-tftp-overlay"
        ][0]
        overlay = [
            action
            for action in prepare.internal_pipeline.actions
            if action.name == "lava-overlay"
        ][0]
        vland = [
            action
            for action in overlay.internal_pipeline.actions
            if action.name == "lava-vland-overlay"
        ][0]
        self.assertTrue(os.path.exists(vland.lava_vland_test_dir))
        vland_files = os.listdir(vland.lava_vland_test_dir)
        self.assertIn("lava-vland-names", vland_files)
        self.assertIn("lava-vland-tags", vland_files)
        self.assertIn("lava-vland-self", vland_files)

    def test_job_no_tags(self):
        with open(self.filename) as yaml_data:
            alpha_data = yaml.safe_load(yaml_data)
        for vlan_key, _ in alpha_data["protocols"][VlandProtocol.name].items():
            alpha_data["protocols"][VlandProtocol.name][vlan_key] = {"tags": []}
        # removed tags from original job to simulate job where any interface tags will be acceptable
        self.assertEqual(
            alpha_data["protocols"][VlandProtocol.name], {"vlan_one": {"tags": []}}
        )
        parser = JobParser()
        job = parser.parse(yaml.dump(alpha_data), self.device, 4212, None, "")
        job.logger = DummyLogger()
        job.validate()
        vprotocol = [
            vprotocol
            for vprotocol in job.protocols
            if vprotocol.name == VlandProtocol.name
        ][0]
        self.assertTrue(vprotocol.valid)
        self.assertEqual(vprotocol.names, {"vlan_one": "4212vlanone"})
        self.assertFalse(vprotocol.check_timeout(120, {"request": "no call"}))
        self.assertRaises(JobError, vprotocol.check_timeout, 60, "deploy_vlans")
        self.assertRaises(
            JobError, vprotocol.check_timeout, 60, {"request": "deploy_vlans"}
        )
        self.assertTrue(vprotocol.check_timeout(120, {"request": "deploy_vlans"}))
        for vlan_name in job.parameters["protocols"][VlandProtocol.name]:
            self.assertIn(vlan_name, vprotocol.params)
            self.assertIn("switch", vprotocol.params[vlan_name])
            self.assertIn("port", vprotocol.params[vlan_name])

    def test_job_bad_tags(self):
        with open(self.filename) as yaml_data:
            alpha_data = yaml.safe_load(yaml_data)
        for vlan_key, _ in alpha_data["protocols"][VlandProtocol.name].items():
            alpha_data["protocols"][VlandProtocol.name][vlan_key] = {
                "tags": ["spurious"]
            }
        # replaced tags from original job to simulate job where an unsupported tag is specified
        self.assertEqual(
            alpha_data["protocols"][VlandProtocol.name],
            {"vlan_one": {"tags": ["spurious"]}},
        )
        parser = JobParser()
        job = parser.parse(yaml.dump(alpha_data), self.device, 4212, None, "")
        job.logger = DummyLogger()
        self.assertRaises(JobError, job.validate)

    def test_primary_interface(self):
        with open(self.filename) as yaml_data:
            alpha_data = yaml.safe_load(yaml_data)
        for interface in self.device["parameters"]["interfaces"]:
            # jinja2 processing of tags: [] results in tags:
            if self.device["parameters"]["interfaces"][interface]["tags"] == []:
                self.device["parameters"]["interfaces"][interface]["tags"] = None
        parser = JobParser()
        job = parser.parse(yaml.dump(alpha_data), self.device, 4212, None, "")
        deploy = [
            action for action in job.pipeline.actions if action.name == "tftp-deploy"
        ][0]
        prepare = [
            action
            for action in deploy.internal_pipeline.actions
            if action.name == "prepare-tftp-overlay"
        ][0]
        overlay = [
            action
            for action in prepare.internal_pipeline.actions
            if action.name == "lava-overlay"
        ][0]
        vland_overlay = [
            action
            for action in overlay.internal_pipeline.actions
            if action.name == "lava-vland-overlay"
        ][0]
        vland_overlay.validate()
        job.logger = DummyLogger()
        job.validate()

    # pylint: disable=protected-access
    def demo(self):
        with open(self.filename) as yaml_data:
            alpha_data = yaml.safe_load(yaml_data)
        vprotocol = VlandProtocol(alpha_data, 422)
        vprotocol.settings = vprotocol.read_settings()
        self.assertIn("port", vprotocol.settings)
        self.assertIn("poll_delay", vprotocol.settings)
        self.assertIn("vland_hostname", vprotocol.settings)
        vprotocol.base_message = {
            "port": vprotocol.settings["port"],
            "poll_delay": vprotocol.settings["poll_delay"],
            "host": vprotocol.settings["vland_hostname"],
            "client_name": socket.gethostname(),
        }
        count = 0
        print("\nTesting vland live using connections.")
        for friendly_name in vprotocol.parameters["protocols"][vprotocol.name]:
            print("Processing VLAN: %s" % friendly_name)
            vprotocol.names[friendly_name] = vprotocol.base_group + "%02d" % count
            count += 1
            vprotocol.vlans[friendly_name], tag = vprotocol._create_vlan(friendly_name)
            print(
                "[%s] Created vlan with id %s"
                % (friendly_name, vprotocol.vlans[friendly_name])
            )
            print("[%s] tag: %s" % (friendly_name, tag))
            for hostname in vprotocol.parameters["protocols"][vprotocol.name][
                friendly_name
            ]:
                params = vprotocol.parameters["protocols"][vprotocol.name][
                    friendly_name
                ][hostname]
                print(
                    "[%s] to use switch %s and port %s"
                    % (friendly_name, params["switch"], params["port"])
                )
                self.assertIn("switch", params)
                self.assertIn("port", params)
                self.assertIsNotNone(params["switch"])
                self.assertIsNotNone(params["port"])
                switch_id = vprotocol._lookup_switch_id(params["switch"])
                self.assertIsNotNone(switch_id)
                print("[%s] Using switch ID %s" % (friendly_name, switch_id))
                port_id = vprotocol._lookup_port_id(switch_id, params["port"])
                print(
                    "%s Looked up port ID %s for %s"
                    % (friendly_name, port_id, params["port"])
                )
                vprotocol._set_port_onto_vlan(vprotocol.vlans[friendly_name], port_id)
                vprotocol.ports.append(port_id)
        print("Finalising - tearing down vlans")
        vprotocol.finalise_protocol()



# Copyright (C) 2014 Linaro Limited
#
# Author: Neil Williams <neil.williams@linaro.org>
#
# This file is part of LAVA Dispatcher.
#
# LAVA Dispatcher is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# LAVA Dispatcher is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along
# with this program; if not, see <http://www.gnu.org/licenses>.

import time
import pytest

from lava_dispatcher.action import Action, Pipeline
from lava_dispatcher.logical import RetryAction
from lava_dispatcher.actions.boot.qemu import BootQEMUImageAction
from lava_dispatcher.actions.test.shell import TestShellRetry
from lava_dispatcher.tests.test_basic import Factory, StdoutTestCase
from lava_dispatcher.tests.test_defs import allow_missing_path


@pytest.fixture
def job():
    return Factory().create_kvm_job("sample_jobs/kvm-repeat.yaml")


def test_basic_structure(job):
    assert job is not None  # nosec - unit test support.
    job.validate()
    assert job.pipeline.errors == []  # nosec - unit test support.
    description_ref = StdoutTestCase.pipeline_reference("kvm-repeat.yaml")
    assert description_ref == job.pipeline.describe(False)  # nosec - unit test support.


def test_deploy_norepeat(job):
    deploy = [deploy for deploy in job.parameters["actions"] if "deploy" in deploy][0][
        "deploy"
    ]
    assert "repeat" not in deploy  # nosec - unit test support.


def test_repeat_yaml(job):
    assert ["repeat"] in [  # nosec - unit test support.
        list(actions.keys()) for actions in job.parameters["actions"]
    ]
    assert "repeat" in job.parameters["actions"][1]  # nosec - unit test support.
    repeat_block = job.parameters["actions"][1]["repeat"]
    assert "count" in repeat_block  # nosec - unit test support.
    # params is a list of default params for the actions, not a list of actions.
    params = [param for param in repeat_block if "count" not in param]
    assert "boot" in repeat_block["actions"][0]  # nosec - unit test support.
    assert "test" in repeat_block["actions"][1]  # nosec - unit test support.
    assert "boot" in repeat_block["actions"][2]  # nosec - unit test support.
    assert "test" in repeat_block["actions"][3]  # nosec - unit test support.
    # count the "namespace" parameter
    assert len(params) == 2  # nosec - unit test support.


def test_nested_structure(job):
    assert ["repeat"] in [  # nosec - unit test support.
        list(actions.keys()) for actions in job.parameters["actions"]
    ]
    # pull out the repeated actions and analyse those
    actions = [retries for retries in job.pipeline.actions if retries.valid]
    assert isinstance(actions[1], BootQEMUImageAction)  # nosec - unit test support.
    assert isinstance(actions[2], TestShellRetry)  # nosec - unit test support.
    assert isinstance(actions[4], TestShellRetry)  # nosec - unit test support.
    assert actions[1].max_retries == 1  # nosec - unit test support.
    assert actions[2].max_retries == 3  # nosec - unit test support.
    assert actions[3].max_retries == 2  # nosec - unit test support.
    assert "repeat-count" in actions[2].parameters  # nosec - unit test support.
    assert (  # nosec - unit test support.
        actions[6].parameters["repeat-count"] > actions[2].parameters["repeat-count"]
    )
    assert (  # nosec - unit test support.
        actions[9].parameters["repeat-count"] > actions[6].parameters["repeat-count"]
    )
    assert (  # nosec - unit test support.
        actions[20].parameters["repeat-count"] > actions[16].parameters["repeat-count"]
    )
    assert (  # nosec - unit test support.
        int([action.level for action in actions if "repeat" in action.parameters][0])
        > 25
    )
    assert "repeat" not in actions[2].parameters  # nosec - unit test support.


def test_single_repeat(job):
    assert ["boot"] in [  # nosec - unit test support.
        list(actions.keys()) for actions in job.parameters["actions"]
    ]
    repeat_actions = [
        action
        for action in job.pipeline.actions
        if isinstance(action, BootQEMUImageAction)
    ]
    boot = repeat_actions[-1]
    assert "repeat" in boot.parameters  # nosec - unit test support.
    assert "repeat-count" not in boot.parameters  # nosec - unit test support.
    repeat_yaml = [
        actions for actions in job.parameters["actions"] if "boot" in actions.keys()
    ][0]["boot"]
    assert "repeat" in repeat_yaml  # nosec - unit test support.
    assert repeat_yaml["repeat"] == 4  # nosec - unit test support.
    assert repeat_yaml["repeat"] == boot.max_retries  # nosec - unit test support.
    assert (  # nosec - unit test support.
        repeat_yaml["repeat"] == job.pipeline.actions[25].parameters["repeat"]
    )
    assert (  # nosec - unit test support.
        "repeat-count" not in job.pipeline.actions[25].parameters
    )


class DummyAction(Action):
    def __init__(self):
        super().__init__()
        self.ran = 0

    def run(self, connection, max_end_time):
        assert connection is None  # nosec - unit test support.
        assert max_end_time == 1  # nosec - unit test support.
        self.ran += 1


def test_repeat_action(monkeypatch):
    monkeypatch.setattr(time, "time", lambda: 0)
    ra = RetryAction()
    ra.parameters = {"repeat": 5}
    ra.level = "1"
    ra.internal_pipeline = Pipeline(parent=ra)
    ra.internal_pipeline.add_action(DummyAction())
    ra.run(None, 1)
    assert ra.internal_pipeline.actions[0].ran == 5  # nosec - unit test support.



# -*- coding: utf-8 -*-
# Generated by Django 1.10.7 on 2017-11-15 09:43
from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [("lava_scheduler_app", "0029_remove_testjob__results_bundle")]

    operations = [
        migrations.RemoveField(model_name="devicetype", name="health_check_job"),
        migrations.RemoveField(model_name="testjob", name="vm_group"),
        migrations.RemoveField(model_name="testjob", name="vmgroup_definition"),
        migrations.RemoveField(model_name="worker", name="is_master"),
        migrations.RemoveField(model_name="worker", name="rpc2_url"),
        migrations.DeleteModel(name="TemporaryDevice"),
    ]



# Copyright (C) 2014 Linaro Limited
#
# Author: Neil Williams <neil.williams@linaro.org>
#
# This file is part of LAVA Dispatcher.
#
# LAVA Dispatcher is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# LAVA Dispatcher is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along
# with this program; if not, see <http://www.gnu.org/licenses>.

import re
import time
import yaml
import decimal
import logging
import pexpect
from nose.tools import nottest
from collections import OrderedDict
from lava_common.exceptions import JobError, InfrastructureError, TestError, LAVABug
from lava_dispatcher.actions.test import TestAction
from lava_dispatcher.action import Pipeline
from lava_dispatcher.logical import LavaTest, RetryAction
from lava_dispatcher.connection import SignalMatch

# pylint: disable=too-many-branches,too-many-statements,too-many-instance-attributes,logging-not-lazy


def handle_testcase(params):
    data = {}
    for param in params:
        parts = param.split("=")
        if len(parts) == 2:
            key, value = parts
            key = key.lower()
            data[key] = value
        else:
            raise JobError('Ignoring malformed parameter for signal: "%s". ' % param)
    return data


class TestShell(LavaTest):
    """
    LavaTestShell Strategy object
    """

    def __init__(self, parent, parameters):
        super().__init__(parent)
        self.action = TestShellRetry()
        self.action.job = self.job
        self.action.section = self.action_type
        parent.add_action(self.action, parameters)

    @classmethod
    def accepts(cls, device, parameters):  # pylint: disable=unused-argument
        if ("definition" in parameters) or ("definitions" in parameters):
            return True, "accepted"
        return False, '"definition" or "definitions" not in parameters'

    @classmethod
    def needs_deployment_data(cls):
        """ Some, not all, deployments will want deployment_data """
        return True

    @classmethod
    def needs_overlay(cls):
        return True

    @classmethod
    def has_shell(cls):
        return True


@nottest
class TestShellRetry(RetryAction):

    name = "lava-test-retry"
    description = "Retry wrapper for lava-test-shell"
    summary = "Retry support for Lava Test Shell"

    def populate(self, parameters):
        self.internal_pipeline = Pipeline(
            parent=self, job=self.job, parameters=parameters
        )
        self.internal_pipeline.add_action(TestShellAction())


# FIXME: move to utils and call inside the overlay
class PatternFixup:
    def __init__(self, testdef, count):
        """
        Like all good arrays, the count is expected to start at zero.
        Avoid calling from validate() or populate() - this needs the
        RepoAction to be running.
        """
        super().__init__()
        self.pat = None
        self.fixup = None
        if isinstance(testdef, dict) and "metadata" in testdef:
            self.testdef = testdef
            self.name = "%d_%s" % (count, testdef["metadata"].get("name"))
        else:
            self.testdef = {}
            self.name = None

    def valid(self):
        return self.fixupdict() and self.pattern() and self.name

    def update(self, pattern, fixupdict):
        if not isinstance(pattern, str):
            raise TestError("Unrecognised test parse pattern type: %s" % type(pattern))
        try:
            self.pat = re.compile(pattern, re.M)
        except re.error as exc:
            raise TestError(
                "Error parsing regular expression %r: %s" % (self.pat, exc.message)
            )
        self.fixup = fixupdict

    def fixupdict(self):
        if "parse" in self.testdef and "fixupdict" in self.testdef["parse"]:
            self.fixup = self.testdef["parse"]["fixupdict"]
        return self.fixup

    def pattern(self):
        if "parse" in self.testdef and "pattern" in self.testdef["parse"]:
            self.pat = self.testdef["parse"]["pattern"]
            if not isinstance(self.pat, str):
                raise TestError(
                    "Unrecognised test parse pattern type: %s" % type(self.pat)
                )
            try:
                self.pat = re.compile(self.pat, re.M)
            except re.error as exc:
                raise TestError(
                    "Error parsing regular expression %r: %s" % (self.pat, exc.message)
                )
        return self.pat


@nottest
class TestShellAction(TestAction):
    """
    Sets up and runs the LAVA Test Shell Definition scripts.
    Supports a pre-command-list of operations necessary on the
    booted image before the test shell can be started.
    """

    name = "lava-test-shell"
    description = "Executing lava-test-runner"
    summary = "Lava Test Shell"

    def __init__(self):
        super().__init__()
        self.signal_director = self.SignalDirector(None)  # no default protocol
        self.patterns = {}
        self.signal_match = SignalMatch()
        self.definition = None
        self.testset_name = None
        self.report = {}
        self.start = None
        self.testdef_dict = {}
        # noinspection PyTypeChecker
        self.pattern = PatternFixup(testdef=None, count=0)
        self.current_run = None

    def _reset_patterns(self):
        # Extend the list of patterns when creating subclasses.
        self.patterns = {
            "exit": "<LAVA_TEST_RUNNER EXIT>",
            "error": "<LAVA_TEST_RUNNER INSTALL_FAIL>",
            "eof": pexpect.EOF,
            "timeout": pexpect.TIMEOUT,
            "signal": r"<LAVA_SIGNAL_(\S+) ([^>]+)>",
        }
        # noinspection PyTypeChecker
        self.pattern = PatternFixup(testdef=None, count=0)

    def validate(self):
        if "definitions" in self.parameters:
            for testdef in self.parameters["definitions"]:
                if "repository" not in testdef:
                    self.errors = "Repository missing from test definition"
        self._reset_patterns()
        super().validate()

    def run(self, connection, max_end_time):  # pylint: disable=too-many-locals
        """
        Common run function for subclasses which define custom patterns
        """
        super().run(connection, max_end_time)

        # Get the connection, specific to this namespace
        connection_namespace = self.parameters.get("connection-namespace")
        parameters = None
        if self.timeout.can_skip(self.parameters):
            self.logger.info(
                "The timeout has 'skip' enabled. "
                "If this test action block times out, the job will continue at the next action block."
            )

        if connection_namespace:
            self.logger.debug("Using connection namespace: %s", connection_namespace)
            parameters = {"namespace": connection_namespace}
        else:
            parameters = {"namespace": self.parameters.get("namespace", "common")}
            self.logger.debug("Using namespace: %s", parameters["namespace"])
        connection = self.get_namespace_data(
            action="shared",
            label="shared",
            key="connection",
            deepcopy=False,
            parameters=parameters,
        )

        if not connection:
            raise LAVABug("No connection retrieved from namespace data")

        self.signal_director.connection = connection

        pattern_dict = {self.pattern.name: self.pattern}
        # pattern dictionary is the lookup from the STARTRUN to the parse pattern.
        self.set_namespace_data(
            action=self.name,
            label=self.name,
            key="pattern_dictionary",
            value=pattern_dict,
        )
        if self.character_delay > 0:
            self.logger.debug(
                "Using a character delay of %i (ms)", self.character_delay
            )

        if not connection.prompt_str:
            connection.prompt_str = [
                self.job.device.get_constant("default-shell-prompt")
            ]
            # FIXME: This should be logged whenever prompt_str is changed, by the connection object.
            self.logger.debug(
                "Setting default test shell prompt %s", connection.prompt_str
            )
        connection.timeout = self.connection_timeout
        # force an initial prompt - not all shells will respond without an excuse.
        connection.sendline(connection.check_char)
        self.wait(connection)

        # use the string instead of self.name so that inheriting classes (like multinode)
        # still pick up the correct command.
        running = self.parameters["stage"]
        pre_command_list = self.get_namespace_data(
            action="test", label="lava-test-shell", key="pre-command-list"
        )
        lava_test_results_dir = self.get_namespace_data(
            action="test", label="results", key="lava_test_results_dir"
        )
        lava_test_sh_cmd = self.get_namespace_data(
            action="test", label="shared", key="lava_test_sh_cmd"
        )

        # Any errors arising from this command are not checked.
        # If the result of the command means that lava-test-runner cannot be found,
        # this will cause the job to time out as Incomplete.
        if pre_command_list:
            for command in pre_command_list:
                connection.sendline(command, delay=self.character_delay)

        if lava_test_results_dir is None:
            raise JobError(
                "Nothing to run. Maybe the 'deploy' stage is missing, "
                "otherwise this is a bug which should be reported."
            )

        self.logger.debug("Using %s" % lava_test_results_dir)
        if lava_test_sh_cmd:
            connection.sendline(
                "export SHELL=%s" % lava_test_sh_cmd, delay=self.character_delay
            )

        try:
            feedbacks = []
            for feedback_ns in self.data.keys():  # pylint: disable=no-member
                if feedback_ns == self.parameters.get("namespace"):
                    continue
                feedback_connection = self.get_namespace_data(
                    action="shared",
                    label="shared",
                    key="connection",
                    deepcopy=False,
                    parameters={"namespace": feedback_ns},
                )
                if feedback_connection:
                    self.logger.debug(
                        "Will listen to feedbacks from '%s' for 1 second", feedback_ns
                    )
                    feedbacks.append((feedback_ns, feedback_connection))

            with connection.test_connection() as test_connection:
                # the structure of lava-test-runner means that there is just one TestAction and it must run all definitions
                test_connection.sendline(
                    "%s/bin/lava-test-runner %s/%s"
                    % (lava_test_results_dir, lava_test_results_dir, running),
                    delay=self.character_delay,
                )

                test_connection.timeout = min(
                    self.timeout.duration, self.connection_timeout.duration
                )
                self.logger.info(
                    "Test shell timeout: %ds (minimum of the action and connection timeout)",
                    test_connection.timeout,
                )

                # Because of the feedbacks, we use a small value for the
                # timeout.  This allows to grab feedback regularly.
                last_check = time.time()
                while self._keep_running(
                    test_connection, test_connection.timeout, connection.check_char
                ):
                    # Only grab the feedbacks every test_connection.timeout
                    if feedbacks and time.time() - last_check > test_connection.timeout:
                        for feedback in feedbacks:
                            # The timeout is really small because the goal is only
                            # to clean the buffer of the feedback connections:
                            # the characters are already in the buffer.
                            # With an higher timeout, this can have a big impact on
                            # the performances of the overall loop.
                            bytes_read = feedback[1].listen_feedback(timeout=1)
                            if bytes_read > 1:
                                self.logger.debug(
                                    "Listened to connection for namespace '%s' done",
                                    feedback[0],
                                )
                        last_check = time.time()
        finally:
            if self.current_run is not None:
                self.logger.error("Marking unfinished test run as failed")
                self.current_run["duration"] = "%.02f" % (time.time() - self.start)
                self.logger.results(self.current_run)  # pylint: disable=no-member
                self.current_run = None

        # Only print if the report is not empty
        if self.report:
            self.logger.debug(yaml.dump(self.report, default_flow_style=False))
        if self.errors:
            raise TestError(self.errors)
        return connection

    def pattern_error(self, test_connection):
        (testrun,) = test_connection.match.groups()
        self.logger.error(
            "Unable to start testrun %s. " "Read the log for more details.", testrun
        )
        self.errors = "Unable to start testrun %s" % testrun
        # This is not accurate but required when exiting.
        self.start = time.time()
        self.current_run = {"definition": "lava", "case": testrun, "result": "fail"}
        return True

    def signal_start_run(self, params):
        self.signal_director.test_uuid = params[1]
        self.definition = params[0]
        uuid = params[1]
        self.start = time.time()
        self.logger.info("Starting test lava.%s (%s)", self.definition, uuid)
        # set the pattern for this run from pattern_dict
        testdef_index = self.get_namespace_data(
            action="test-definition", label="test-definition", key="testdef_index"
        )
        uuid_list = self.get_namespace_data(
            action="repo-action", label="repo-action", key="uuid-list"
        )
        for (key, value) in enumerate(testdef_index):
            if self.definition == "%s_%s" % (key, value):
                pattern_dict = self.get_namespace_data(
                    action="test", label=uuid_list[key], key="testdef_pattern"
                )
                if not pattern_dict:
                    self.logger.info("Skipping test definition patterns.")
                    continue
                pattern = pattern_dict["testdef_pattern"]["pattern"]
                fixup = pattern_dict["testdef_pattern"]["fixupdict"]
                self.patterns.update({"test_case_result": re.compile(pattern, re.M)})
                self.pattern.update(pattern, fixup)
                self.logger.info("Enabling test definition pattern %r" % pattern)
                self.logger.info(
                    "Enabling test definition fixup %r" % self.pattern.fixup
                )
        self.current_run = {
            "definition": "lava",
            "case": self.definition,
            "uuid": uuid,
            "result": "fail",
        }
        testdef_commit = self.get_namespace_data(
            action="test", label=uuid, key="commit-id"
        )
        if testdef_commit:
            self.current_run.update({"commit_id": testdef_commit})

    def signal_end_run(self, params):
        self.definition = params[0]
        uuid = params[1]
        # remove the pattern for this run from pattern_dict
        self._reset_patterns()
        # catch error in ENDRUN being handled without STARTRUN
        if not self.start:
            self.start = time.time()
        self.logger.info("Ending use of test pattern.")
        self.logger.info(
            "Ending test lava.%s (%s), duration %.02f",
            self.definition,
            uuid,
            time.time() - self.start,
        )
        self.current_run = None
        res = {
            "definition": "lava",
            "case": self.definition,
            "uuid": uuid,
            "repository": self.get_namespace_data(
                action="test", label=uuid, key="repository"
            ),
            "path": self.get_namespace_data(action="test", label=uuid, key="path"),
            "duration": "%.02f" % (time.time() - self.start),
            "result": "pass",
        }
        revision = self.get_namespace_data(action="test", label=uuid, key="revision")
        res["revision"] = revision if revision else "unspecified"
        res["namespace"] = self.parameters["namespace"]
        connection_namespace = self.parameters.get("connection_namespace")
        if connection_namespace:
            res["connection-namespace"] = connection_namespace
        commit_id = self.get_namespace_data(action="test", label=uuid, key="commit-id")
        if commit_id:
            res["commit_id"] = commit_id

        self.logger.results(res)  # pylint: disable=no-member
        self.start = None

    @nottest
    def signal_test_case(self, params):
        # If the STARTRUN signal was not received correctly, we cannot continue
        # as the test_uuid is missing.
        # This is only happening when the signal string is split by some kernel messages.
        if self.signal_director.test_uuid is None:
            self.logger.error(
                "Unknown test uuid. The STARTRUN signal for this test action was not received correctly."
            )
            raise TestError("Invalid TESTCASE signal")
        try:
            data = handle_testcase(params)
            # get the fixup from the pattern_dict
            res = self.signal_match.match(data, fixupdict=self.pattern.fixupdict())
        except (JobError, TestError) as exc:
            self.logger.error(str(exc))
            return True

        p_res = self.get_namespace_data(
            action="test", label=self.signal_director.test_uuid, key="results"
        )
        if not p_res:
            p_res = OrderedDict()
            self.set_namespace_data(
                action="test",
                label=self.signal_director.test_uuid,
                key="results",
                value=p_res,
            )

        # prevent losing data in the update
        # FIXME: support parameters and retries
        if res["test_case_id"] in p_res:
            raise JobError(
                "Duplicate test_case_id in results: %s" % res["test_case_id"]
            )
        # turn the result dict inside out to get the unique
        # test_case_id/testset_name as key and result as value
        res_data = {
            "definition": self.definition,
            "case": res["test_case_id"],
            "result": res["result"],
        }
        # check for measurements
        if "measurement" in res:
            try:
                measurement = decimal.Decimal(res["measurement"])
            except decimal.InvalidOperation:
                raise TestError("Invalid measurement %s" % res["measurement"])
            res_data["measurement"] = measurement
            if "units" in res:
                res_data["units"] = res["units"]

        if self.testset_name:
            res_data["set"] = self.testset_name
            self.report[res["test_case_id"]] = {
                "set": self.testset_name,
                "result": res["result"],
            }
        else:
            self.report[res["test_case_id"]] = res["result"]
        # Send the results back
        self.logger.results(res_data)  # pylint: disable=no-member

    @nottest
    def signal_test_reference(self, params):
        if len(params) != 3:
            raise TestError("Invalid use of TESTREFERENCE")
        res_dict = {
            "case": params[0],
            "definition": self.definition,
            "result": params[1],
            "reference": params[2],
        }
        if self.testset_name:
            res_dict.update({"set": self.testset_name})
        self.logger.results(res_dict)  # pylint: disable=no-member

    @nottest
    def signal_test_feedback(self, params):
        feedback_ns = params[0]
        if feedback_ns not in self.data.keys():
            self.logger.error("%s is not a valid namespace")
            return
        self.logger.info("Requesting feedback from namespace: %s", feedback_ns)
        feedback_connection = self.get_namespace_data(
            action="shared",
            label="shared",
            key="connection",
            deepcopy=False,
            parameters={"namespace": feedback_ns},
        )
        feedback_connection.listen_feedback(timeout=1)

    @nottest
    def signal_test_set(self, params):
        name = None
        action = params.pop(0)
        if action == "START":
            name = "testset_" + action.lower()
            try:
                self.testset_name = params[0]
            except IndexError:
                raise JobError("Test set declared without a name")
            self.logger.info("Starting test_set %s", self.testset_name)
        elif action == "STOP":
            self.logger.info("Closing test_set %s", self.testset_name)
            self.testset_name = None
            name = "testset_" + action.lower()
        return name

    @nottest
    def pattern_test_case(self, test_connection):
        match = test_connection.match
        if match is pexpect.TIMEOUT:
            self.logger.warning("err: lava_test_shell has timed out (test_case)")
            return False
        res = self.signal_match.match(
            match.groupdict(), fixupdict=self.pattern.fixupdict()
        )
        self.logger.debug("outer_loop_result: %s" % res)
        return True

    @nottest
    def pattern_test_case_result(self, test_connection):
        res = test_connection.match.groupdict()
        fixupdict = self.pattern.fixupdict()
        if fixupdict and res["result"] in fixupdict:
            res["result"] = fixupdict[res["result"]]
        if res:
            # disallow whitespace in test_case_id
            test_case_id = "%s" % res["test_case_id"].replace("/", "_")
            self.logger.marker({"case": res["test_case_id"], "type": "test_case"})
            if " " in test_case_id.strip():
                self.logger.debug(
                    "Skipping invalid test_case_id '%s'", test_case_id.strip()
                )
                return True
            res_data = {
                "definition": self.definition,
                "case": res["test_case_id"],
                "result": res["result"],
            }
            # check for measurements
            if "measurement" in res:
                try:
                    measurement = decimal.Decimal(res["measurement"])
                except decimal.InvalidOperation:
                    raise TestError("Invalid measurement %s" % res["measurement"])
                res_data["measurement"] = measurement
                if "units" in res:
                    res_data["units"] = res["units"]

            self.logger.results(res_data)  # pylint: disable=no-member
            self.report[res["test_case_id"]] = res["result"]
        return True

    def check_patterns(
        self, event, test_connection, check_char
    ):  # pylint: disable=unused-argument
        """
        Defines the base set of pattern responses.
        Stores the results of testcases inside the TestAction
        Call from subclasses before checking subclass-specific events.
        """
        ret_val = False
        if event == "exit":
            self.logger.info("ok: lava_test_shell seems to have completed")
            self.testset_name = None

        elif event == "error":
            # Parsing is not finished
            ret_val = self.pattern_error(test_connection)

        elif event == "eof":
            self.testset_name = None
            raise InfrastructureError("lava_test_shell connection dropped.")

        elif event == "timeout":
            # allow feedback in long runs
            ret_val = True

        elif event == "signal":
            name, params = test_connection.match.groups()
            self.logger.debug("Received signal: <%s> %s" % (name, params))
            params = params.split()
            if name == "STARTRUN":
                self.signal_start_run(params)
            elif name == "ENDRUN":
                self.signal_end_run(params)
            elif name == "STARTTC":
                self.logger.marker({"case": params[0], "type": "start_test_case"})
            elif name == "ENDTC":
                self.logger.marker({"case": params[0], "type": "end_test_case"})
            elif name == "TESTCASE":
                self.logger.marker(
                    {
                        "case": params[0].replace("TEST_CASE_ID=", ""),
                        "type": "test_case",
                    }
                )
                self.signal_test_case(params)
            elif name == "TESTFEEDBACK":
                self.signal_test_feedback(params)
            elif name == "TESTREFERENCE":
                self.signal_test_reference(params)
            elif name == "TESTSET":
                ret = self.signal_test_set(params)
                if ret:
                    name = ret
            elif name == "TESTRAISE":
                raise TestError(" ".join(params))
            elif name == "TESTEVENT":
                self.logger.event(" ".join(params))

            self.signal_director.signal(name, params)
            ret_val = True

        elif event == "test_case":
            ret_val = self.pattern_test_case(test_connection)
        elif event == "test_case_result":
            ret_val = self.pattern_test_case_result(test_connection)
        return ret_val

    def _keep_running(self, test_connection, timeout, check_char):
        if "test_case_results" in self.patterns:
            self.logger.info(
                "Test case result pattern: %r" % self.patterns["test_case_results"]
            )
        retval = test_connection.expect(list(self.patterns.values()), timeout=timeout)
        return self.check_patterns(
            list(self.patterns.keys())[retval], test_connection, check_char
        )

    class SignalDirector:

        # FIXME: create proxy handlers
        def __init__(self, protocol=None):
            """
            Base SignalDirector for singlenode jobs.
            MultiNode and LMP jobs need to create a suitable derived class as both also require
            changes equivalent to the old _keep_running functionality.

            SignalDirector is the link between the Action and the Connection. The Action uses
            the SignalDirector to interact with the I/O over the Connection.
            """
            self.protocol = protocol  # communicate externally over the protocol API
            self.connection = None  # communicate with the device
            self.logger = logging.getLogger("dispatcher")
            self.test_uuid = None

        def setup(self, parameters):
            """
            Allows the parent Action to pass extra data to a customised SignalDirector
            """
            pass

        def signal(self, name, params):
            handler = getattr(self, "_on_" + name.lower(), None)
            if handler:
                try:
                    # The alternative here is to drop the getattr and have a long if:elif:elif:else.
                    # Without python support for switch, this gets harder to read than using
                    # a getattr lookup for the callable (codehelp). So disable checkers:
                    # noinspection PyCallingNonCallable
                    handler(*params)
                except TypeError as exc:
                    # handle serial corruption which can overlap kernel messages onto test output.
                    self.logger.exception(str(exc))
                    raise TestError(
                        "Unable to handle the test shell signal correctly: %s"
                        % str(exc)
                    )
                except JobError as exc:
                    self.logger.error(
                        "job error: handling signal %s failed: %s", name, exc
                    )
                    return False
                return True



# -*- coding: utf-8 -*-
# Copyright (C) 2015-2018 Linaro Limited
#
# Author: Neil Williams <neil.williams@linaro.org>
#         Remi Duraffort <remi.duraffort@linaro.org>
#
# This file is part of LAVA.
#
# LAVA is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License version 3
# as published by the Free Software Foundation
#
# LAVA is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with LAVA.  If not, see <http://www.gnu.org/licenses/>.

"""
Database utility functions which use but are not actually models themselves
Used to allow models.py to be shortened and easier to follow.
"""

# pylint: disable=wrong-import-order

import os
import yaml
import jinja2
import json
import logging
from nose.tools import nottest

from django.db.models import Q, Case, When, IntegerField, Sum
from django.contrib.auth.models import User
from django.core.exceptions import ValidationError
from django.core.validators import validate_email

from lava_scheduler_app.models import (
    Device,
    DeviceType,
    NotificationRecipient,
    TestJob,
    Worker,
)
from lava_scheduler_app.schema import validate_submission, SubmissionException
from lava_results_app.dbutils import map_metadata
from lava_results_app.models import Query

# pylint: disable=too-many-branches,too-many-statements,too-many-locals


def match_vlan_interface(device, job_def):
    if not isinstance(job_def, dict):
        raise RuntimeError("Invalid vlan interface data")
    if (
        "protocols" not in job_def
        or "lava-vland" not in job_def["protocols"]
        or not device
    ):
        return False
    interfaces = []
    logger = logging.getLogger("lava-master")
    device_dict = device.load_configuration()
    if not device_dict or device_dict.get("parameters", {}).get("interfaces") is None:
        return False

    for vlan_name in job_def["protocols"]["lava-vland"]:
        tag_list = job_def["protocols"]["lava-vland"][vlan_name]["tags"]
        for interface in device_dict["parameters"]["interfaces"]:
            tags = device_dict["parameters"]["interfaces"][interface]["tags"]
            if not tags:
                continue
            logger.info(
                "Job requests %s for %s, device %s provides %s for %s",
                tag_list,
                vlan_name,
                device.hostname,
                tags,
                interface,
            )
            if (
                set(tags) & set(tag_list) == set(tag_list)
                and interface not in interfaces
            ):
                logger.info(
                    "Matched vlan %s to interface %s on %s",
                    vlan_name,
                    interface,
                    device,
                )
                interfaces.append(interface)
                # matched, do not check any further interfaces of this device for this vlan
                break

    logger.info(
        "Matched: %s",
        (len(interfaces) == len(job_def["protocols"]["lava-vland"].keys())),
    )
    return len(interfaces) == len(job_def["protocols"]["lava-vland"].keys())


# TODO: check the list of exception that can be raised
@nottest
def testjob_submission(job_definition, user, original_job=None):
    """
    Single submission frontend for YAML
    :param job_definition: string of the job submission
    :param user: user attempting the submission
    :return: a job or a list of jobs
    :raises: SubmissionException, Device.DoesNotExist,
        DeviceType.DoesNotExist, DevicesUnavailableException,
        ValueError
    """
    json_data = True
    try:
        # accept JSON but store as YAML
        json.loads(job_definition)
    except json.decoder.JSONDecodeError:
        json_data = False
    if json_data:
        # explicitly convert to YAML.
        # JSON cannot have comments anyway.
        job_definition = yaml.safe_dump(yaml.safe_load(job_definition))

    validate_job(job_definition)
    # returns a single job or a list (not a QuerySet) of job objects.
    job = TestJob.from_yaml_and_user(job_definition, user, original_job=original_job)
    return job


def parse_job_description(job):
    filename = os.path.join(job.output_dir, "description.yaml")
    logger = logging.getLogger("lava-master")
    try:
        with open(filename, "r") as f_describe:
            description = f_describe.read()
        pipeline = yaml.load(description)
    except (OSError, yaml.YAMLError):
        logger.error("'Unable to open and parse '%s'", filename)
        return

    if not map_metadata(description, job):
        logger.warning("[%d] unable to map metadata", job.id)

    # add the compatibility result from the master to the definition for comparison on the slave.
    try:
        compat = int(pipeline["compatibility"])
    except (TypeError, ValueError):
        compat = pipeline["compatibility"] if pipeline is not None else None
        logger.error("[%d] Unable to parse job compatibility: %s", job.id, compat)
        compat = 0
    job.pipeline_compatibility = compat
    job.save(update_fields=["pipeline_compatibility"])


def device_type_summary(visible=None):
    devices = (
        Device.objects.filter(
            ~Q(health=Device.HEALTH_RETIRED) & Q(device_type__in=visible)
        )
        .only("state", "health", "is_public", "device_type", "hostname")
        .values("device_type")
        .annotate(
            idle=Sum(
                Case(
                    When(
                        state=Device.STATE_IDLE,
                        health__in=[Device.HEALTH_GOOD, Device.HEALTH_UNKNOWN],
                        worker_host__state=Worker.STATE_ONLINE,
                        then=1,
                    ),
                    default=0,
                    output_field=IntegerField(),
                )
            ),
            busy=Sum(
                Case(
                    When(
                        state__in=[Device.STATE_RESERVED, Device.STATE_RUNNING], then=1
                    ),
                    default=0,
                    output_field=IntegerField(),
                )
            ),
            offline=Sum(
                Case(
                    When(
                        Q(state=Device.STATE_IDLE)
                        & (
                            Q(worker_host__state=Worker.STATE_OFFLINE)
                            | ~Q(health__in=[Device.HEALTH_GOOD, Device.HEALTH_UNKNOWN])
                        ),
                        then=1,
                    ),
                    default=0,
                    output_field=IntegerField(),
                )
            ),
            restricted=Sum(
                Case(
                    When(is_public=False, then=1),
                    default=0,
                    output_field=IntegerField(),
                )
            ),
        )
        .order_by("device_type")
    )
    return devices


def active_device_types():
    """
    Filter the available device types to exclude
    all device-types where ALL devices are in health RETIRED
    without excluding device-types where only SOME devices are retired.

    oneliner:
{device.device_type for device in Device.objects.filter(~Q(health=Device.HEALTH_RETIRED))}.union( \
{dt for dt in {device.device_type for device in Device.objects.filter(health=Device.HEALTH_RETIRED)} \
if list(Device.objects.filter(Q(device_type=dt), ~Q(health=Device.HEALTH_RETIRED)))})

    Returns a RestrictedQuerySet of DeviceType objects.
    """
    not_retired_devices = Device.objects.filter(
        Q(device_type__display=True), ~Q(health=Device.HEALTH_RETIRED)
    ).select_related("device_type")
    retired_devices = Device.objects.filter(
        Q(device_type__display=True), health=Device.HEALTH_RETIRED
    ).select_related("device_type")
    not_all_retired = (
        set()
    )  # set of device_type.names where some devices of that device_type are retired but *not* all.
    for device in retired_devices:
        # identify device_types which can be added back because not all devices of that type are retired.
        if list(
            Device.objects.filter(
                Q(device_type=device.device_type), ~Q(health=Device.HEALTH_RETIRED)
            )
        ):
            not_all_retired.add(device.device_type.name)
    # join the two sets as a union.
    candidates = {device.device_type.name for device in not_retired_devices}.union(
        not_all_retired
    )
    device_types = DeviceType.objects.filter(name__in=candidates)
    return device_types


def load_devicetype_template(device_type_name, raw=False):
    """
    Loads the bare device-type template as a python dictionary object for
    representation within the device_type templates.
    No device-specific details are parsed - default values only, so some
    parts of the dictionary may be unexpectedly empty. Not to be used when
    rendering device configuration for a testjob.
    :param device_type_name: DeviceType.name (string)
    :param raw: if True, return the raw yaml
    :return: None or a dictionary of the device type template.
    """
    path = os.path.dirname(Device.CONFIG_PATH)
    type_loader = jinja2.FileSystemLoader([os.path.join(path, "device-types")])
    env = jinja2.Environment(  # nosec - YAML, not HTML, no XSS scope.
        loader=jinja2.ChoiceLoader([type_loader]), trim_blocks=True, autoescape=False
    )
    try:
        template = env.get_template("%s.jinja2" % device_type_name)
        data = template.render()
        if not data:
            return None
        return data if raw else yaml.safe_load(data)
    except (jinja2.TemplateError, yaml.error.YAMLError):
        return None


def invalid_template(dt):
    """
    Careful with the inverted logic here.
    Return True if the template is invalid.
    See unit tests in test_device.py
    """
    d_template = bool(
        load_devicetype_template(dt.name)
    )  # returns None on error ( == False)
    if not d_template:
        queryset = list(
            Device.objects.filter(Q(device_type=dt), ~Q(health=Device.HEALTH_RETIRED))
        )
        if not queryset:
            return False
        extends = set([device.get_extends() for device in queryset])
        if not extends:
            return True
        for extend in extends:
            if not extend:
                return True
            d_template = not bool(
                load_devicetype_template(extend.replace(".jinja2", ""))
            )
            # if d_template is False, template is valid, invalid_template returns False
            if d_template:
                return True
    else:
        d_template = False  # template exists, invalid check is False
    return d_template


def validate_job(data):
    try:
        yaml_data = yaml.safe_load(data)
    except yaml.YAMLError as exc:
        raise SubmissionException("Loading job submission failed: %s." % exc)

    # validate against the submission schema.
    validate_submission(yaml_data)  # raises SubmissionException if invalid.
    validate_yaml(yaml_data)  # raises SubmissionException if invalid.


def validate_yaml(yaml_data):
    if "notify" in yaml_data:
        if "recipients" in yaml_data["notify"]:
            for recipient in yaml_data["notify"]["recipients"]:
                if recipient["to"]["method"] == NotificationRecipient.EMAIL_STR:
                    if "email" not in recipient["to"] and "user" not in recipient["to"]:
                        raise SubmissionException(
                            "No valid user or email address specified."
                        )
                else:
                    if (
                        "handle" not in recipient["to"]
                        and "user" not in recipient["to"]
                    ):
                        raise SubmissionException(
                            "No valid user or IRC handle specified."
                        )
                if "user" in recipient["to"]:
                    try:
                        User.objects.get(username=recipient["to"]["user"])
                    except User.DoesNotExist:
                        raise SubmissionException(
                            "%r is not an existing user in LAVA."
                            % recipient["to"]["user"]
                        )
                elif "email" in recipient["to"]:
                    try:
                        validate_email(recipient["to"]["email"])
                    except ValidationError:
                        raise SubmissionException(
                            "%r is not a valid email address."
                            % recipient["to"]["email"]
                        )

        if (
            "compare" in yaml_data["notify"]
            and "query" in yaml_data["notify"]["compare"]
        ):
            query_yaml_data = yaml_data["notify"]["compare"]["query"]
            if "username" in query_yaml_data:
                try:
                    query = Query.objects.get(
                        owner__username=query_yaml_data["username"],
                        name=query_yaml_data["name"],
                    )
                    if query.content_type.model_class() != TestJob:
                        raise SubmissionException("Only TestJob queries allowed.")
                except Query.DoesNotExist:
                    raise SubmissionException(
                        "Query ~%s/%s does not exist"
                        % (query_yaml_data["username"], query_yaml_data["name"])
                    )
            else:  # Custom query.
                if query_yaml_data["entity"] != "testjob":
                    raise SubmissionException("Only TestJob queries allowed.")
                try:
                    conditions = None
                    if "conditions" in query_yaml_data:
                        conditions = query_yaml_data["conditions"]
                    Query.validate_custom_query(query_yaml_data["entity"], conditions)
                except Exception as e:
                    raise SubmissionException(e)



# -*- coding: utf-8 -*-
# Generated by Django 1.9 on 2015-12-07 09:21
from __future__ import unicode_literals

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('dashboard_app', '0020_auto_20151124_1419'),
    ]

    operations = [
        migrations.AlterOrderWithRespectTo(
            name='testresult',
            order_with_respect_to=None,
        ),
    ]



import os
import yaml

# pylint: disable=superfluous-parens,ungrouped-imports
from lava_scheduler_app.schema import validate_device
from lava_scheduler_app.tests.test_base_templates import (
    BaseTemplate,
    prepare_jinja_template,
)

# pylint: disable=too-many-branches,too-many-public-methods
# pylint: disable=too-many-nested-blocks


class TestFastbootTemplates(BaseTemplate.BaseTemplateCases):
    """
    Test rendering of jinja2 templates

    When adding or modifying a jinja2 template, add or update the test here.
    Use realistic data - complete exports of the device dictionary preferably.
    Set debug to True to see the content of the rendered templates
    Set system to True to use the system templates - note that this requires
    that the templates in question are in sync with the branch upon which the
    test is run. Therefore, if the templates should be the same, this can be
    used to check that the templates are correct. If there are problems, check
    for a template with a .dpkg-dist extension. Check the diff between the
    checkout and the system file matches the difference between the system file
    and the dpkg-dist version. If the diffs match, copy the dpkg-dist onto the
    system file.
    """

    def test_nexus4_template(self):
        data = """{% extends 'nexus4.jinja2' %}
{% set adb_serial_number = 'R32D300FRYP' %}
{% set fastboot_serial_number = 'R32D300FRYP' %}
{% set device_info = [{'board_id': 'R32D300FRYP'}] %}
{% set static_info = [{'board_id': 'R32D300FRYP'}] %}
"""
        self.assertTrue(self.validate_data("nexus4-01", data))
        template_dict = prepare_jinja_template("nexus4-01", data, raw=False)
        self.assertEqual("R32D300FRYP", template_dict["adb_serial_number"])
        self.assertEqual("R32D300FRYP", template_dict["fastboot_serial_number"])
        self.assertEqual([], template_dict["fastboot_options"])
        self.assertIsNotNone(template_dict)
        self.assertIsInstance(template_dict["device_info"], list)
        self.assertIsInstance(template_dict["static_info"], list)

    def test_x15_template(self):
        data = """{% extends 'x15.jinja2' %}
{% set adb_serial_number = '1234567890' %}
{% set fastboot_serial_number = '1234567890' %}
{% set interrupt_prompt = "interrupt bootloader" %}
"""
        self.assertTrue(self.validate_data("x15-01", data))
        template_dict = prepare_jinja_template("x15-01", data, raw=False)
        self.assertIsInstance(template_dict["adb_serial_number"], str)
        self.assertEqual("1234567890", template_dict["adb_serial_number"])
        self.assertEqual("1234567890", template_dict["fastboot_serial_number"])
        self.assertEqual([], template_dict["fastboot_options"])
        self.assertIn("u-boot", template_dict["actions"]["boot"]["methods"])
        self.assertIn(
            "parameters", template_dict["actions"]["boot"]["methods"]["u-boot"]
        )
        self.assertIn(
            "interrupt_prompt",
            template_dict["actions"]["boot"]["methods"]["u-boot"]["parameters"],
        )
        self.assertEqual(
            "interrupt bootloader",
            template_dict["actions"]["boot"]["methods"]["u-boot"]["parameters"][
                "interrupt_prompt"
            ],
        )
        # fastboot deploy to eMMC
        self.assertIn("mmc", template_dict["actions"]["boot"]["methods"]["u-boot"])
        self.assertIn(
            "commands", template_dict["actions"]["boot"]["methods"]["u-boot"]["mmc"]
        )
        # NFS using standard U-Boot TFTP
        self.assertIn("nfs", template_dict["actions"]["boot"]["methods"]["u-boot"])
        self.assertIn(
            "commands", template_dict["actions"]["boot"]["methods"]["u-boot"]["nfs"]
        )
        for command in template_dict["actions"]["boot"]["methods"]["u-boot"]["nfs"][
            "commands"
        ]:
            if "setenv bootargs" in command:
                # x15 needs both consoles enabled.
                self.assertIn("ttyS2", command)
                self.assertNotIn("console=ttyO2", command)
        for board in template_dict["device_info"]:
            self.assertEqual(template_dict["fastboot_serial_number"], board["board_id"])

    def test_sharkl2_template(self):
        data = """{% extends 'sharkl2.jinja2' %}
{% set adb_serial_number = 'R32D300FRYP' %}
{% set fastboot_serial_number = 'R32D300FRYP' %}
"""
        self.assertTrue(self.validate_data("sharkl2-01", data))
        template_dict = prepare_jinja_template("sharkl2-01", data, raw=False)
        self.assertEqual("R32D300FRYP", template_dict["adb_serial_number"])
        self.assertEqual("R32D300FRYP", template_dict["fastboot_serial_number"])
        self.assertEqual([], template_dict["fastboot_options"])
        self.assertIn("u-boot", template_dict["actions"]["boot"]["methods"])
        self.assertIn(
            "parameters", template_dict["actions"]["boot"]["methods"]["u-boot"]
        )
        self.assertIn(
            "interrupt_prompt",
            template_dict["actions"]["boot"]["methods"]["u-boot"]["parameters"],
        )
        # fastboot boot off eMMC
        self.assertIn("fastboot", template_dict["actions"]["boot"]["methods"])
        self.assertIn("reboot", template_dict["actions"]["boot"]["methods"]["fastboot"])

    def test_nexus10_template(self):
        data = """{% extends 'nexus10.jinja2' %}
{% set adb_serial_number = 'R32D300FRYP' %}
{% set fastboot_serial_number = 'R32D300FRYP' %}
{% set soft_reboot_command = 'adb -s R32D300FRYP reboot bootloader' %}
{% set connection_command = 'adb -s R32D300FRYP shell' %}
{% set device_info = [{'board_id': 'R32D300FRYP'}] %}
{% set static_info = [{'board_id': 'R32D300FRYP'}] %}
"""
        self.assertTrue(self.validate_data("staging-nexus10-01", data))
        template_dict = prepare_jinja_template("staging-pixel-01", data, raw=False)
        self.assertIsNotNone(template_dict)
        self.assertIsInstance(template_dict["device_info"], list)
        self.assertIsInstance(template_dict["static_info"], list)

    def test_hikey_r2_template(self):
        with open(
            os.path.join(
                os.path.dirname(__file__), "devices", "hi6220-hikey-r2-01.jinja2"
            )
        ) as hikey:
            data = hikey.read()
        self.assertIsNotNone(data)
        self.assertTrue(self.validate_data("hi6220-hikey-01", data))
        template_dict = prepare_jinja_template("staging-hikey-01", data, raw=False)
        self.assertIsNotNone(template_dict)
        self.assertIsInstance(template_dict["device_info"], list)
        self.assertIsInstance(template_dict["static_info"], list)
        self.assertEqual(template_dict["device_info"][0]["board_id"], "0123456789")
        self.assertEqual(template_dict["static_info"][0]["board_id"], "S_N0123456")
        self.assertIsInstance(template_dict["fastboot_options"], list)
        self.assertEqual(template_dict["fastboot_options"], ["-S", "256M"])
        order = template_dict["flash_cmds_order"]
        self.assertEqual(0, order.index("ptable"))
        self.assertEqual(1, order.index("xloader"))
        self.assertEqual(2, order.index("fastboot"))

    def test_hikey_template(self):
        with open(
            os.path.join(os.path.dirname(__file__), "devices", "hi6220-hikey-01.jinja2")
        ) as hikey:
            data = hikey.read()
        self.assertIsNotNone(data)
        self.assertTrue(self.validate_data("hi6220-hikey-01", data))
        template_dict = prepare_jinja_template("staging-hikey-01", data, raw=False)
        self.assertIsNotNone(template_dict)
        self.assertIsInstance(template_dict["device_info"], list)
        self.assertIsInstance(template_dict["static_info"], list)
        self.assertEqual(template_dict["device_info"][0]["board_id"], "0123456789")
        self.assertEqual(template_dict["static_info"][0]["board_id"], "S_N0123456")
        self.assertIsInstance(template_dict["fastboot_options"], list)
        self.assertEqual(template_dict["fastboot_options"], ["-S", "256M"])
        order = template_dict["flash_cmds_order"]
        self.assertEqual(0, order.index("ptable"))
        self.assertEqual(1, order.index("fastboot"))
        self.assertIn("cache", order)
        self.assertIn("system", order)
        self.assertIn("userdata", order)
        self.assertNotIn("xloader", order)
        # test support for retreiving MAC from device.
        data += "{% set device_mac = '00:E0:4C:53:44:58' %}"
        self.assertTrue(self.validate_data("hi6220-hikey-01", data))
        template_dict = prepare_jinja_template("staging-hikey-01", data, raw=False)
        self.assertIn("parameters", template_dict)
        self.assertIn("interfaces", template_dict["parameters"])
        self.assertIn("target", template_dict["parameters"]["interfaces"])
        self.assertIn("mac", template_dict["parameters"]["interfaces"]["target"])
        self.assertIn("ip", template_dict["parameters"]["interfaces"]["target"])
        self.assertIsNotNone(template_dict["parameters"]["interfaces"]["target"]["mac"])
        self.assertNotEqual(
            "", template_dict["parameters"]["interfaces"]["target"]["mac"]
        )
        self.assertIsNone(template_dict["parameters"]["interfaces"]["target"]["ip"])

    def test_hikey_grub_efi(self):
        with open(
            os.path.join(os.path.dirname(__file__), "devices", "hi6220-hikey-01.jinja2")
        ) as hikey:
            data = hikey.read()
        self.assertIsNotNone(data)
        job_ctx = {"kernel": "Image", "devicetree": "hi6220-hikey.dtb"}
        self.assertTrue(self.validate_data("hi6220-hikey-01", data))
        template_dict = prepare_jinja_template(
            "staging-hikey-01", data, job_ctx=job_ctx, raw=False
        )
        self.assertIsNotNone(template_dict)
        self.assertIsNotNone(template_dict["actions"]["boot"]["methods"])
        self.assertIn("grub-efi", template_dict["actions"]["boot"]["methods"])
        self.assertEqual(
            "fastboot",
            template_dict["actions"]["boot"]["methods"]["grub-efi"]["menu_options"],
        )
        params = template_dict["actions"]["boot"]["methods"]["grub-efi"]
        self.assertEqual(params["parameters"]["bootloader_prompt"], "grub>")
        for command in params["installed"]["commands"]:
            if command.startswith("search"):
                self.assertIn("rootfs", command)
            elif command.startswith("linux"):
                self.assertIn("/boot/Image", command)
            elif command.startswith("devicetree"):
                self.assertIn("hi6220-hikey.dtb", command)
            elif "root=" in command:
                self.assertIn("/dev/mmcblk0p9", command)
                self.assertIn("ttyAMA3", command)
            else:
                self.assertEqual("boot", command)
        self.assertIn("ssh", template_dict["actions"]["deploy"]["methods"])
        params = template_dict["actions"]["deploy"]["methods"]["ssh"]
        self.assertIsNotNone(params)
        self.assertIn("port", params)
        self.assertIn("user", params)
        self.assertIn("options", params)
        self.assertIn("identity_file", params)

    def test_hikey620_uarts(self):
        with open(
            os.path.join(os.path.dirname(__file__), "devices", "hi6220-hikey-01.jinja2")
        ) as hikey:
            data = hikey.read()
        self.assertIsNotNone(data)
        job_ctx = {}
        self.assertTrue(self.validate_data("hi6220-hikey-01", data))
        template_dict = prepare_jinja_template(
            "staging-hikey-01", data, job_ctx=job_ctx, raw=False
        )
        validate_device(template_dict)
        self.assertIsNotNone(template_dict)
        self.assertIn("commands", template_dict)
        self.assertNotIn("connect", template_dict["commands"])
        self.assertIn("connections", template_dict["commands"])
        self.assertIn("uart0", template_dict["commands"]["connections"])
        self.assertIn("uart1", template_dict["commands"]["connections"])
        self.assertIn("tags", template_dict["commands"]["connections"]["uart1"])
        self.assertIn(
            "primary", template_dict["commands"]["connections"]["uart1"]["tags"]
        )
        self.assertNotIn("tags", template_dict["commands"]["connections"]["uart0"])
        self.assertEqual(
            template_dict["commands"]["connections"]["uart0"]["connect"],
            "telnet localhost 4002",
        )
        self.assertEqual(
            template_dict["commands"]["connections"]["uart1"]["connect"],
            "telnet 192.168.1.200 8001",
        )

    def test_hikey960_grub(self):
        with open(
            os.path.join(os.path.dirname(__file__), "devices", "hi960-hikey-01.jinja2")
        ) as hikey:
            data = hikey.read()
        self.assertIsNotNone(data)
        job_ctx = {"kernel": "Image", "devicetree": "hi960-hikey.dtb"}
        # self.assertTrue(self.validate_data('hi960-hikey-01', data))
        template_dict = prepare_jinja_template(
            "staging-hikey-01", data, job_ctx=job_ctx, raw=False
        )
        self.assertIsNotNone(template_dict)
        self.assertIsNotNone(template_dict["actions"]["boot"]["methods"])
        self.assertNotIn(
            "menu_options", template_dict["actions"]["boot"]["methods"]["grub"]
        )
        self.assertIn("grub", template_dict["actions"]["boot"]["methods"])
        params = template_dict["actions"]["boot"]["methods"]["grub"]
        for command in params["installed"]["commands"]:
            self.assertEqual("boot", command)
        self.assertIn("ssh", template_dict["actions"]["deploy"]["methods"])
        params = template_dict["actions"]["deploy"]["methods"]["ssh"]
        self.assertIsNotNone(params)
        self.assertIn("port", params)
        self.assertIn("user", params)
        self.assertIn("options", params)
        self.assertIn("identity_file", params)

        # test support for retreiving MAC from device using base-fastboot.
        data += "{% set device_mac = '00:E0:4C:53:44:58' %}"
        self.assertTrue(self.validate_data("hi6220-hikey-01", data))
        template_dict = prepare_jinja_template(
            "staging-hikey-01", data, job_ctx=None, raw=False
        )
        self.assertIn("parameters", template_dict)
        self.assertIn("interfaces", template_dict["parameters"])
        self.assertIn("target", template_dict["parameters"]["interfaces"])
        self.assertIn("mac", template_dict["parameters"]["interfaces"]["target"])
        self.assertIn("ip", template_dict["parameters"]["interfaces"]["target"])
        self.assertIsNotNone(template_dict["parameters"]["interfaces"]["target"]["mac"])
        self.assertNotEqual(
            "", template_dict["parameters"]["interfaces"]["target"]["mac"]
        )
        self.assertEqual(
            "00:E0:4C:53:44:58",
            template_dict["parameters"]["interfaces"]["target"]["mac"],
        )
        self.assertIsNone(template_dict["parameters"]["interfaces"]["target"]["ip"])

    def test_nexus5x_template(self):
        data = """{% extends 'nexus5x.jinja2' %}
{% set adb_serial_number = '10de1214adae123' %}
{% set fastboot_serial_number = '10de1214adae123' %}
{% set device_info = [{'board_id': '10de1214adae123'}] %}
{% set static_info = [{'board_id': '10de1214adae123'}] %}
"""
        self.assertTrue(self.validate_data("staging-nexus5x-01", data))
        template_dict = prepare_jinja_template("staging-pixel-01", data, raw=False)
        self.assertIsNotNone(template_dict)
        self.assertIsInstance(template_dict["device_info"], list)
        self.assertIsInstance(template_dict["static_info"], list)

    def test_pixel_template(self):
        data = """{% extends 'pixel.jinja2' %}
{% set adb_serial_number = 'FDAC1231DAD' %}
{% set fastboot_serial_number = 'FDAC1231DAD' %}
{% set device_info = [{'board_id': 'FDAC1231DAD'}] %}
{% set static_info = [{'board_id': 'FDAC1231DAD'}] %}
"""
        self.assertTrue(self.validate_data("staging-pixel-01", data))
        template_dict = prepare_jinja_template("staging-pixel-01", data, raw=False)
        self.assertIsNotNone(template_dict)
        self.assertIsInstance(template_dict["device_info"], list)
        self.assertIsInstance(template_dict["static_info"], list)

    def test_nuc_template(self):
        self.assertTrue(
            self.validate_data(
                "staging-nuc-01",
                """{% extends 'adb-nuc.jinja2' %}
{% set device_ip = '192.168.1.11' %}
""",
            )
        )

    def test_ifc6410(self):
        data = """{% extends 'ifc6410.jinja2' %}
{% set adb_serial_number = 'e080c212' %}
{% set fastboot_serial_number = 'e080c212' %}
        """
        self.assertTrue(self.validate_data("staging-ifc6410-01", data))
        template_dict = prepare_jinja_template("staging-ifc6410-01", data, raw=False)
        self.assertEqual("e080c212", template_dict["adb_serial_number"])
        self.assertEqual("e080c212", template_dict["fastboot_serial_number"])
        self.assertEqual([], template_dict["fastboot_options"])
        methods = template_dict["actions"]["boot"]["methods"]["fastboot"]
        self.assertIn("reboot", methods)
        self.assertIn("boot", methods)
        self.assertIn("auto-login", methods)
        self.assertIn("overlay-unpack", methods)

    def test_db820c_template(self):
        data = """{% extends 'dragonboard-820c.jinja2' %}
{% set adb_serial_number = '3083f595' %}
{% set fastboot_serial_number = '3083f595' %}
"""
        self.assertTrue(self.validate_data("db820c-01", data))
        template_dict = prepare_jinja_template("db820c-01", data, raw=False)
        self.assertEqual("3083f595", template_dict["adb_serial_number"])
        self.assertEqual("3083f595", template_dict["fastboot_serial_number"])
        self.assertEqual([], template_dict["fastboot_options"])

    def test_recovery_mode(self):
        with open(
            os.path.join(
                os.path.dirname(__file__), "devices", "hi6220-hikey-bl-01.jinja2"
            )
        ) as hikey:
            data = hikey.read()
        self.assertTrue(self.validate_data("hi620-bl-01", data))
        template_dict = prepare_jinja_template("hi620-bl-01", data, raw=False)
        recovery = template_dict["actions"]["deploy"]["methods"]
        self.assertIsNotNone("recovery", recovery)
        self.assertIn("recovery", recovery)
        self.assertIn("commands", recovery["recovery"])
        self.assertIsNotNone("recovery", recovery["recovery"]["commands"])
        self.assertIn("recovery_mode", recovery["recovery"]["commands"])
        self.assertEqual(
            [
                "/home/neil/lava-lab/shared/lab-scripts/eth008_control -a 10.15.0.171 -r 1 -s off",
                "/home/neil/lava-lab/shared/lab-scripts/eth008_control -a 10.15.0.171 -r 2 -s on",
            ],
            recovery["recovery"]["commands"]["recovery_mode"],
        )
        self.assertIn("recovery_exit", recovery["recovery"]["commands"])
        self.assertEqual(
            [
                "/home/neil/lava-lab/shared/lab-scripts/eth008_control -a 10.15.0.171 -r 1 -s on",
                "/home/neil/lava-lab/shared/lab-scripts/eth008_control -a 10.15.0.171 -r 2 -s off",
            ],
            recovery["recovery"]["commands"]["recovery_exit"],
        )

    def test_sdm845_mtp_template(self):
        fastboot_cmd_order = [
            "update",
            "ptable",
            "partition",
            "hyp",
            "modem",
            "rpm",
            "sbl1",
            "sbl2",
            "sec",
            "tz",
            "aboot",
            "cdt",
            "boot",
            "rootfs",
            "vendor",
            "system",
            "cache",
            "userdata",
        ]

        rendered = self.render_device_dictionary_file("sdm845-mtp-05.jinja2")
        template_dict = yaml.safe_load(rendered)
        self.assertEqual("5c302cef", template_dict["adb_serial_number"])
        self.assertEqual("5c302cef", template_dict["fastboot_serial_number"])
        self.assertEqual([], template_dict["fastboot_options"])

        for cmd in template_dict["flash_cmds_order"]:
            idx = template_dict["flash_cmds_order"].index(cmd)
            self.assertEqual(cmd, fastboot_cmd_order[idx])



# Copyright (C) 2018 Linaro Limited
#
# Author: Remi Duraffort <remi.duraffort@linaro.org>
#
# This file is part of LAVA Dispatcher.
#
# LAVA Dispatcher is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# LAVA Dispatcher is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along
# with this program; if not, see <http://www.gnu.org/licenses>.

import shlex
import yaml

from lava_dispatcher.action import Pipeline
from lava_dispatcher.actions.deploy import DeployAction
from lava_dispatcher.actions.deploy.download import DownloaderAction
from lava_dispatcher.actions.deploy.environment import DeployDeviceEnvironment
from lava_dispatcher.actions.deploy.overlay import OverlayAction
from lava_dispatcher.logical import Deployment, RetryAction
from lava_dispatcher.utils.strings import substitute


class FlasherRetryAction(RetryAction):

    name = "deploy-flasher-retry"
    description = "deploy flasher with retry"
    summary = "deploy custom flasher"

    def populate(self, parameters):
        self.internal_pipeline = Pipeline(
            parent=self, job=self.job, parameters=parameters
        )
        self.internal_pipeline.add_action(FlasherAction())


class FlasherAction(DeployAction):

    name = "deploy-flasher"
    description = "deploy flasher"
    summary = "deploy custom flasher"

    def __init__(self):
        super().__init__()
        self.commands = []
        self.path = None

    def validate(self):
        super().validate()
        method = self.job.device["actions"]["deploy"]["methods"]["flasher"]
        self.commands = method.get("commands")
        if not isinstance(self.commands, list):
            self.errors = "'commands' should be a list"

    def populate(self, parameters):
        self.internal_pipeline = Pipeline(
            parent=self, job=self.job, parameters=parameters
        )

        # Download the images
        self.path = self.mkdtemp()
        for image in parameters["images"].keys():
            self.internal_pipeline.add_action(DownloaderAction(image, self.path))

        if self.test_needs_deployment(parameters):
            self.internal_pipeline.add_action(DeployDeviceEnvironment())
        if self.test_needs_overlay(parameters):
            self.internal_pipeline.add_action(OverlayAction())

    def run(self, connection, max_end_time):
        connection = super().run(connection, max_end_time)
        # Substitute in the device commands
        substitutions = {}
        for key in self.parameters["images"].keys():
            filename = self.get_namespace_data(
                action="download-action", label=key, key="file"
            )
            filename = filename[len(self.path) + 1 :]
            substitutions["{%s}" % key.upper()] = filename

        # Add power commands
        substitutions["{HARD_RESET_COMMAND}"] = self.job.device.hard_reset_command
        substitutions["{SOFT_RESET_COMMAND}"] = self.job.device.soft_reset_command
        substitutions["{PRE_OS_COMMAND}"] = self.job.device.pre_os_command
        if substitutions["{PRE_OS_COMMAND}"] is None:
            substitutions["{PRE_OS_COMMAND}"] = ""
        substitutions["{PRE_POWER_COMMAND}"] = self.job.device.pre_power_command
        if substitutions["{PRE_POWER_COMMAND}"] is None:
            substitutions["{PRE_POWER_COMMAND}"] = ""
        substitutions["{POWER_ON_COMMAND}"] = self.job.device.power_command
        substitutions["{POWER_OFF_COMMAND}"] = self.job.device.get("commands", {}).get(
            "power_off", ""
        )

        # Add some device configuration
        substitutions["{DEVICE_INFO}"] = yaml.dump(
            self.job.device.get("device_info", [])
        )
        substitutions["{STATIC_INFO}"] = yaml.dump(
            self.job.device.get("static_info", [])
        )

        # Run the commands
        for cmd in self.commands:
            cmds = substitute(shlex.split(cmd), substitutions)
            self.run_cmd(cmds, error_msg="Unable to flash the device", cwd=self.path)

        return connection


class Flasher(Deployment):
    compatibility = 4
    name = "flasher"

    def __init__(self, parent, parameters):
        super().__init__(parent)
        self.action = FlasherRetryAction()
        self.action.section = self.action_type
        self.action.job = self.job
        parent.add_action(self.action, parameters)

    @classmethod
    def accepts(cls, device, parameters):
        if "flasher" not in device["actions"]["deploy"]["methods"]:
            return False, "'flasher' not in the device configuration deploy methods"
        if parameters["to"] != "flasher":
            return False, '"to" parameter is not "flasher"'
        return True, "accepted"



# -*- coding: utf-8 -*-
# Copyright (C) 2015-2018 Linaro Limited
#
# Author: Stevan Radakovic <stevan.radakovic@linaro.org>
#
# This file is part of LAVA.
#
# LAVA is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License version 3
# as published by the Free Software Foundation
#
# LAVA is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with LAVA.  If not, see <http://www.gnu.org/licenses/>.

from django.core.exceptions import PermissionDenied
from django.shortcuts import get_object_or_404
from django.utils.decorators import available_attrs
from django.utils.functional import wraps

from lava_results_app.models import Query


def ownership_required(view_func):
    @wraps(view_func, assigned=available_attrs(view_func))
    def wrapper(request, *args, **kwargs):
        report_name = kwargs.get("name")
        username = kwargs.get("username")
        query = get_object_or_404(Query, name=report_name, owner__username=username)
        if query.is_accessible_by(request.user):
            return view_func(request, *args, **kwargs)
        else:
            raise PermissionDenied

    return wrapper



#  Copyright 2013-2014 Linaro Limited
#  Author Neil Williams <neil.williams@linaro.org>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.


import logging
import json
import uuid
from nose.tools import nottest


# disable pylint warnings until lava-coordinator is updated to make it easier to port other changes.
# lava-coordinator would need to be repackaged to make a module package and support a non-socket mode.
# pylint: disable=superfluous-parens,invalid-name,logging-not-lazy,attribute-defined-outside-init,no-self-use


class TestSignals:

    message_str = ""

    def formatString(self, reply):
        if isinstance(reply, dict):
            for target, messages in reply.items():
                for key, value in messages.items():
                    self.message_str += " %s:%s=%s" % (target, key, value)
        return self.message_str

    def checkMessage(self, reply):
        if reply is not None:
            self.log = logging.getLogger("testCase")
            self.log.info("\t<LAVA_TEST_COMPLETE%s>" % self.formatString(reply))


class TestSocket:
    """
    Mock up a LavaCoordinator socket without needing a server
    """

    response = None
    header = True
    log = None
    message = None
    passes = 0
    signalHandler = None

    def __init__(self):
        self.log = logging.getLogger("testCase")
        self.signalHandler = TestSignals()
        self.response = None

    def send(self, data):
        if self.header:
            self.header = False
            assert int(data, 16) < 0xFFFE  # nosec - unit test support
            self.log.info("\tCoordinator header: %d bytes" % int(data, 16))
        else:
            try:
                json_data = json.loads(data)
            except ValueError:
                assert False  # nosec - unit test support
            if not self.response:
                assert json_data["response"] == "nack"  # nosec - unit test support
                self.header = True
                return
            assert "response" in json_data  # nosec - unit test support
            self.log.info("\tCoordinator response: '%s'" % json_data["response"])
            self.log.info("\tdebug: %s" % self.response["response"])
            assert (  # nosec - unit test support
                json_data["response"] == self.response["response"]
            )
            self.passes += 1
            if self.message:
                # we are expecting a message back.
                assert "message" in json_data  # nosec - unit test support
                self.log.info(
                    "\tCoordinator received a message: '%s'"
                    % (json.dumps(json_data["message"]))
                )
                assert json_data["message"] == self.message  # nosec - unit test support
                self.passes += 1
            else:
                # actual calls will discriminate between dict and string replies
                # according to the call prototype itself
                if "message" in json_data:
                    if isinstance(json_data["message"], dict):
                        self.log.info(
                            "\tCould have expected a message: '%s'"
                            % json.dumps(json_data["message"])
                        )
                    else:
                        self.log.info("\t<LAVA_TEST_REPLY %s>" % json_data["message"])
                self.passes += 1
            self.header = True
            self.response = json_data

    def get_response(self):
        """ Without a socket, need to get the actual response direct
        """
        return self.response

    def close(self):
        self.log.info("\tCoordinator closing.")

    def clearPasses(self):
        self.passes = 0

    def logPasses(self):
        if self.passes == 1:
            self.log.info("\tCoordinator: %d socket test passed" % self.passes)
        else:
            self.log.info("\tCoordinator: %d socket tests passed" % self.passes)

    def prepare(self, name):
        if name:
            self.response = {"response": name}
        if self.response:
            self.log.info(
                "\tCoordinator: expecting a response: '%s'" % self.response["response"]
            )

    def validate(self, message):
        self.message = message
        if self.message:
            self.log.info(
                "\tCoordinator: expecting a message: '%s'" % json.dumps(self.message)
            )
        self.signalHandler.checkMessage(self.message)


@nottest
class TestCoordinator:
    """
    Contains direct copies of critical functions from lava.coordinator module in a wrapper
    which prevents the need for socket communication. Module imported from lava-coordinator 0.1.5
    """

    running = True
    json_data = None
    group_name = None
    group_size = 0
    client_name = None
    conn = None
    log = None
    rpc_delay = 2
    blocksize = 4 * 1024
    host = "localhost"
    all_groups = {}

    def __init__(self):
        self.group_name = str(uuid.uuid4())
        self.conn = TestSocket()
        self.log = logging.getLogger("testCase")
        self.log.info("")
        self.json_data = {"request": "testing"}
        self.client_name = "testpoller"
        self.log.info(
            "\tStarting test with %s %d %d %s"
            % (json.dumps(self.json_data), self.rpc_delay, self.blocksize, self.host)
        )
        self.expectResponse(None)

    def _updateData(self, json_data):
        """
        Sanity checks the JSON data and retrieves the data for the group specified
        :param json_data: JSON request
        :return: the client_name specified in the JSON to allow the message handler to lookup
        the correct messages within this group.
        """
        self._clear_group()
        if "client_name" in json_data:
            client_name = json_data["client_name"]
        else:
            raise RuntimeError("Missing client_name in request: %s" % json_data)
        if json_data["group_name"] not in self.all_groups:
            if "group_size" not in json_data or json_data["group_size"] == 0:
                logging.error(
                    "%s asked for a new group %s without specifying the size of the group"
                    % (client_name, json_data["group_name"])
                )
                return None
            # auto register a new group
            self.group["count"] = int(json_data["group_size"])
            self.group["group"] = json_data["group_name"]
            self.all_groups[json_data["group_name"]] = self.group
            logging.info(
                "The %s group will contain %d nodes."
                % (self.group["group"], self.group["count"])
            )
        self.group = self.all_groups[json_data["group_name"]]
        # now add this client to the registered data for this group
        if client_name not in self.group["clients"]:
            self.group["clients"][client_name] = json_data["hostname"]
            if json_data["role"] not in self.group["roles"]:
                self.group["roles"][json_data["role"]] = []
            self.group["roles"][json_data["role"]].append(client_name)
        return client_name

    def _clear_group(self):
        self.group = {
            "group": "",
            "count": 0,
            "complete": 0,
            "rpc_delay": self.rpc_delay,
            "clients": {},
            "roles": {},
            "syncs": {},
            "messages": {},
            "waits": {},
            "bundles": {},
        }

    def _clearGroupData(self, json_data):
        """
        Clears the group data once all nodes have finished.
        Nodes do *not* wait for other nodes to finish.
        :param json_data: incoming JSON request
        """
        if "group_name" not in json_data:
            self._badRequest()
            return
        if json_data["group_name"] not in self.all_groups:
            self._badRequest()
            return
        self.group["complete"] += 1
        logging.debug(
            "clear Group Data: %d of %d"
            % (self.group["complete"], len(self.group["clients"]))
        )
        self._ackResponse()
        if len(self.group["clients"]) > self.group["complete"]:
            return
        logging.debug("Clearing group data for %s" % json_data["group_name"])
        del self.all_groups[json_data["group_name"]]
        self._clear_group()

    def _setGroupData(self, json_data):
        """
        Implements the wait until all clients in this group have connected
        :rtype : None
        :param json_data: incoming JSON request
        """
        if len(self.group["clients"]) != self.group["count"]:
            logging.info(
                "Waiting for %d more clients to connect to %s group"
                % (
                    (
                        self.group["count"] - len(self.group["clients"]),
                        json_data["group_name"],
                    )
                )
            )
            # group_data is not complete yet.
            self._waitResponse()
            return
        logging.info("Group complete, starting tests")
        # client_name must be unique because it's the DB index & conf file name
        group_data = {}
        for role in self.group["roles"]:
            for client in self.group["roles"][role]:
                group_data[client] = role
        msg = {"response": "group_data", "roles": group_data}
        msgdata = self._formatMessage(msg)
        if msgdata:
            self.conn.send(msgdata[0])
            self.conn.send(msgdata[1])
        self.conn.close()

    def _formatMessage(self, message):
        """ Prepares the LAVA Coordinator header and a JSON string
        of the message ready for transmission. Currently, the
        header is just the length of the JSON string as a hexadecimal
        string padded to 8 characters (not including 0x)
        :param message: Python object suitable for conversion into JSON
        :rtype : A tuple - first value is the header, second value is
        the data to send, returns None if the message could not be formatted.
        """
        try:
            msgstr = json.dumps(message)
        except ValueError:
            return None
        # "header" calculation
        msglen = "%08X" % len(msgstr)
        if int(msglen, 16) > 0xFFFFFFFF:
            logging.error(
                "Message was too long to send! %d > %d" % (int(msglen, 16), 0xFFFFFFFF)
            )
            return None
        return msglen, msgstr

    def _sendMessage(self, client_name, messageID):
        """ Sends a message to the currently connected client.
        (the "connection name" or hostname of the connected client does not necessarily
        match the name of the client registered with the group.)
        :param client_name: the client_name to lookup for the message
        :param messageID: the message index set by lavaSend
        :rtype : None
        """
        if (
            client_name not in self.group["messages"]
            or messageID not in self.group["messages"][client_name]
        ):
            logging.error(
                "Unable to find messageID %s for client %s" % (messageID, client_name)
            )
            self._badRequest()
            return
        logging.info(
            "Sending messageID '%s' to %s in group %s: %s"
            % (
                messageID,
                client_name,
                self.group["group"],
                json.dumps(self.group["messages"][client_name][messageID]),
            )
        )
        msg = {
            "response": "ack",
            "message": self.group["messages"][client_name][messageID],
        }
        msgdata = self._formatMessage(msg)
        if msgdata:
            logging.info(
                "Sending response to %s in group %s: %s"
                % (client_name, self.group["group"], json.dumps(msg))
            )
            self.conn.send(msgdata[0])
            self.conn.send(msgdata[1])
        self.conn.close()

    def _sendWaitMessage(self, client_name, messageID):
        """ Sends a wait message to the currently connected client.
        (the "connection name" or hostname of the connected client does not necessarily
        match the name of the client registered with the group.)
        :param client_name: the client_name to lookup for the message
        :param messageID: the message index set by lavaSend
        :rtype : None
        """
        if (
            messageID not in self.group["waits"]
            or client_name not in self.group["waits"][messageID]
        ):
            logging.error(
                "Unable to find messageID %s for client %s" % (messageID, client_name)
            )
            self._badRequest()
            return
        logging.info(
            "Sending wait messageID '%s' to %s in group %s: %s"
            % (
                messageID,
                client_name,
                self.group["group"],
                json.dumps(self.group["waits"][messageID]["data"]),
            )
        )
        msg = {"response": "ack", "message": self.group["waits"][messageID]["data"]}
        msgdata = self._formatMessage(msg)
        if msgdata:
            logging.info(
                "Sending wait response to %s in group %s: %s"
                % (client_name, self.group["group"], json.dumps(msg))
            )
            self.conn.send(msgdata[0])
            self.conn.send(msgdata[1])
        self.conn.close()

    def _getMessage(self, json_data):
        # message value is allowed to be None as long as the message key exists.
        if "message" not in json_data:
            return {}
        if "messageID" not in json_data:
            logging.error(
                "No 'messageID' key found in request %s when looking for message."
                % json.dumps(json_data)
            )
            return {}
        if json_data["message"] is None:
            return {}
        return json_data["message"]

    def _getMessageID(self, json_data):
        if "messageID" not in json_data:
            logging.error(
                "No 'messageID' key found in request %s when looking for ID"
                % json.dumps(json_data)
            )
            return None
        return json_data["messageID"]

    def _badRequest(self):
        msgdata = self._formatMessage({"response": "nack"})
        if msgdata:
            self.conn.send(msgdata[0])
            self.conn.send(msgdata[1])
        self.conn.close()

    def _ackResponse(self):
        msgdata = self._formatMessage({"response": "ack"})
        if msgdata:
            self.conn.send(msgdata[0])
            self.conn.send(msgdata[1])
        self.conn.close()

    def _waitResponse(self):
        msgdata = self._formatMessage({"response": "wait"})
        if msgdata:
            self.conn.send(msgdata[0])
            self.conn.send(msgdata[1])
        self.conn.close()

    def lavaSync(self, json_data, client_name):
        """
        Global synchronization primitive. Sends a message and waits for the same
        message from all of the other devices.
        """
        logging.debug(
            "Coordinator:lavaSync %s from %s in group %s"
            % (json.dumps(json_data), client_name, self.group["group"])
        )
        messageID = self._getMessageID(json_data)
        message = self._getMessage(json_data)
        # send the messageID as the message if message is empty
        if not message:
            message = messageID
        logging.info(
            "LavaSync request for '%s' at stage '%s' in group '%s'"
            % (client_name, messageID, self.group["group"])
        )
        self.group["syncs"].setdefault(messageID, {})
        self.group["messages"].setdefault(client_name, {}).setdefault(messageID, {})
        if len(self.group["syncs"][messageID]) >= self.group["count"]:
            self.group["messages"][client_name][messageID] = message
            self._sendMessage(client_name, messageID)
            # mark this client as having picked up the message
            self.group["syncs"][messageID][client_name] = 0
        else:
            logging.info(
                "waiting for '%s': not all clients in group '%s' have been seen yet %d < %d"
                % (
                    messageID,
                    self.group["group"],
                    len(self.group["syncs"][messageID]),
                    self.group["count"],
                )
            )
            self.group["messages"][client_name][messageID] = message
            self.group["syncs"][messageID][client_name] = 1
            self._waitResponse()
            return
        # clear the sync data for this messageID when the last client connects to
        # allow the message to be re-used later for another sync
        clear_syncs = True
        for pending in self.group["syncs"][messageID]:
            if self.group["syncs"][messageID][pending]:
                clear_syncs = False
        if clear_syncs:
            logging.debug(
                "Clearing all sync messages for '%s' in group '%s'"
                % (messageID, self.group["group"])
            )
            self.group["syncs"][messageID].clear()

    def lavaWaitAll(self, json_data, client_name):
        """
        Waits until all other devices in the group send a message with the given message ID.
        IF <role> is passed, only wait until all devices with that given role send a message.
        """
        logging.debug("lavaWaitAll:json_data: %s" % json_data)
        messageID = self._getMessageID(json_data)
        if "waitrole" in json_data:
            for client in self.group["roles"][json_data["role"]]:
                if messageID not in self.group["waits"]:
                    logging.debug("messageID %s not yet seen" % messageID)
                    self._waitResponse()
                    return
                if client not in self.group["waits"][messageID]:
                    # FIXME: bug? if this client has not sent the messageID yet,
                    # causing it to wait will simply force a timeout. node needs
                    # to output a warning, so maybe send a "nack" ?
                    self._waitResponse()
                    return
                if client in self.group["waits"]:
                    logging.debug(
                        "replying: %s for %s"
                        % (self.group["waits"][client][messageID], client)
                    )
            if client_name in self.group["waits"]:
                logging.debug(
                    "lavaWaitAll message: %s"
                    % json.dumps(self.group["waits"][client_name][messageID])
                )
        else:
            for client in self.group["clients"]:
                logging.debug("checking %s for wait message" % client)
                if messageID not in self.group["waits"]:
                    logging.debug("messageID %s not yet seen" % messageID)
                    self._waitResponse()
                    return
                if client not in self.group["waits"][messageID]:
                    logging.debug("setting waiting for %s" % client)
                    self._waitResponse()
                    return
        self._sendWaitMessage(client_name, messageID)

    def lavaWait(self, json_data, client_name):
        """
        Waits until any other device in the group sends a message with the given ID.
        This call will block the client until such message is sent, the server continues.
        :param json_data: the JSON request
        :param client_name: the client_name to receive the message
        """
        messageID = self._getMessageID(json_data)
        if (
            client_name not in self.group["messages"]
            or messageID not in self.group["messages"][client_name]
        ):
            logging.debug("MessageID %s not yet seen for %s" % (messageID, client_name))
            self._waitResponse()
            return
        self._sendMessage(client_name, messageID)

    def lavaSend(self, json_data, client_name):
        """
        A message list won't be seen by the destination until the destination
        calls lava_wait or lava_wait_all with the messageID
        If lava_wait is called first, the message will be sent when the client reconnects
        messages are broadcast - picked up by lava-wait or lava-sync - any call to lava-wait will pick up
            the complete message.
        waits are not broadcast - only picked up by lava-wait-all - all calls to lava-wait-all will
            wait until all clients have used lava-send for the same messageID
        """
        message = self._getMessage(json_data)
        messageID = self._getMessageID(json_data)
        logging.info(
            "lavaSend handler in Coordinator received a messageID '%s' for group '%s' from %s"
            % (messageID, self.group["group"], client_name)
        )
        if client_name not in self.group["messages"]:
            self.group["messages"][client_name] = {}
        # construct the message hash which stores the data from each client separately
        # but which gets returned as a complete hash upon request
        msg_hash = {}
        msg_hash.update({client_name: message})
        # always set this client data if the call is made to update the broadcast
        if messageID not in self.group["messages"][client_name]:
            self.group["messages"][client_name][messageID] = {}
        self.group["messages"][client_name][messageID].update(msg_hash)
        logging.debug(
            "message %s for %s"
            % (json.dumps(self.group["messages"][client_name][messageID]), client_name)
        )
        # now broadcast the message into the other clients in this group
        for client in self.group["clients"]:
            if client not in self.group["messages"]:
                self.group["messages"][client] = {}
            if messageID not in self.group["messages"][client]:
                self.group["messages"][client][messageID] = {}
            self.group["messages"][client][messageID].update(msg_hash)
            logging.debug(
                "broadcast %s for %s"
                % (json.dumps(self.group["messages"][client][messageID]), client)
            )
        # separate the waits from the messages for wait-all support
        if messageID not in self.group["waits"]:
            self.group["waits"][messageID] = {}
        if client_name not in self.group["waits"][messageID]:
            self.group["waits"][messageID][client_name] = {}
        if "data" not in self.group["waits"][messageID]:
            self.group["waits"][messageID]["data"] = {}
        self.group["waits"][messageID]["data"].update(msg_hash)
        self._ackResponse()

    def dataReceived(self, json_data):
        """
        Handles all incoming data for the singleton LAVA Coordinator
        :param json_data: the incoming data stream - expected to be JSON
        """
        logging.debug("data=%s" % json.dumps(json_data))
        if "request" not in json_data:
            self._badRequest()
            return
        request = json_data["request"]
        # retrieve the group data for the group which contains this client and get the client name
        # self-register using the group_size, if necessary
        client_name = self._updateData(json_data)
        if not client_name or not self.group["group"]:
            logging.info("no client_name or group found")
            raise RuntimeError("no client_name or group found")
            # return
        if request == "group_data":
            self._setGroupData(json_data)
        elif request == "clear_group":
            self._clearGroupData(json_data)
        # elif request == "aggregate":
        #     logging.debug("Aggregate called")
        #     self._aggregateBundle(json_data, client_name)
        elif request == "lava_sync":
            logging.debug(
                "lava_sync: %s request made by '%s' in group '%s'"
                % (json.dumps(json_data), client_name, self.group["group"])
            )
            self.lavaSync(json_data, client_name)
        elif request == "lava_wait_all":
            logging.debug("lava_wait_all: %s" % json_data)
            self.lavaWaitAll(json_data, client_name)
        elif request == "lava_wait":
            logging.debug("lava_wait: %s" % json_data)
            self.lavaWait(json_data, client_name)
        elif request == "lava_send":
            logging.info("lava_send: %s" % json_data)
            self.lavaSend(json_data, client_name)
        elif request == "complete":
            logging.info(
                "coordinator communication for '%s' in group '%s' is complete, closing."
                % (client_name, self.group["group"])
            )
            self.conn.close()
        else:
            logging.error("Unrecognised request %s. Closed connection." % json_data)
            self._badRequest()
        return self.conn.get_response()

    def newGroup(self, size):
        self.group_name = str(uuid.uuid4())
        self.group_size = size
        self.log = logging.getLogger("testCase")
        self.log.info("\tGroup name %s" % self.group_name)

    # sets up TestSocket for the correct assertions
    def expectResponse(self, test_name):
        self.conn.prepare(test_name)

    def expectMessage(self, message):
        self.conn.validate(message)

    def addClient(self, client_name):
        self.conn.response = "ack"
        self.client_name = client_name
        self.log = logging.getLogger("testCase")
        ret = self._updateData(
            {
                "client_name": client_name,
                "group_size": self.group_size,
                "role": "tester",
                "hostname": "localhost",
                "group_name": self.group_name,
            }
        )
        self.log.info(
            "\tAdded client_name '%s'. group size now: %d"
            % (client_name, len(self.group["clients"]))
        )
        self.log.info("\tCurrent client_name: '%s'" % self.client_name)
        return ret

    def addClientRole(self, client_name, role):
        self.conn.response = "ack"
        self.client_name = client_name
        self.log = logging.getLogger("testCase")
        ret = self._updateData(
            {
                "client_name": client_name,
                "group_size": self.group_size,
                "role": role,
                "hostname": "localhost",
                "group_name": self.group_name,
            }
        )
        self.log.info(
            "\tAdded client_name '%s' with role '%s'. group size now: %d"
            % (client_name, role, len(self.group["clients"]))
        )
        self.log.info("\tCurrent client_name: '%s'" % self.client_name)
        return ret
